# -------------------------------------------------------------------------- #
# Copyright 2010-2011, University of Chicago                                 #
#                                                                            #
# Licensed under the Apache License, Version 2.0 (the "License"); you may    #
# not use this file except in compliance with the License. You may obtain    #
# a copy of the License at                                                   #
#                                                                            #
# http://www.apache.org/licenses/LICENSE-2.0                                 #
#                                                                            #
# Unless required by applicable law or agreed to in writing, software        #
# distributed under the License is distributed on an "AS IS" BASIS,          #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.   #
# See the License for the specific language governing permissions and        #
# limitations under the License.                                             #
# -------------------------------------------------------------------------- #
"""
The Globus Provision API

This API is currently meant to be used locally (i.e., not accessible through a 
network via a remote call interface like REST, SOAP, etc.) and by a single 
user (i.e., there is no notion of different users owning different instances). 
Future versions of Globus Provision may run as a daemon with a remotely-accessible 
API that supports multiple users.
"""

import sys
import traceback
import os.path

from boto.exception import EC2ResponseError

import globus.provision.deploy.ec2 as ec2_deploy
import globus.provision.deploy.dummy as dummy_deploy

from globus.provision.core.deploy import DeploymentException
from globus.provision.core.topology import Topology, Node
from globus.provision.core.instance import InstanceStore, InstanceException
from globus.provision.common.threads import MultiThread
from globus.provision.common.ssh import SSHCommandFailureException
from globus.provision.common import log
from globus.provision.common.config import ConfigException
from globus.provision.common.persistence import ObjectValidationException,\
    PropertyChange
from globus.provision.common.go_transfer import GlobusOnlineCLIHelper, GlobusOnlineHelper,\
    GlobusOnlineException

class API(object):
    """
    The Globus Provision API.
    """
    
    STATUS_SUCCESS = 0
    STATUS_FAIL = 1
    
    def __init__(self, instances_dir):
        self.instances_dir = instances_dir
        
    def instance_create(self, topology_json, config_txt):
        try:
            istore = InstanceStore(self.instances_dir)
            inst = istore.create_new_instance(topology_json, config_txt)
        except ConfigException, cfge:
            message = "Error in configuration file: %s" % cfge
            return (API.STATUS_FAIL, message, None)
        except ObjectValidationException, ove:
            message = "Error in topology file: %s" % ove
            return (API.STATUS_FAIL, message, None)        
        except:
            message = self.__unexpected_exception_to_text(what = "creating a new instance.")
            return (API.STATUS_FAIL, message, None)
        
        return (API.STATUS_SUCCESS, "Success", inst.id)

    def instance(self, inst_id):
        (success, message, inst) = self.__get_instance(inst_id)
        
        if not success:
            return (API.STATUS_FAIL, message, None)
                
        try:
            json_string = inst.topology.to_json_string()
        except:
            message = self.__unexpected_exception_to_text(what = "accessing the instance's topology.")
            return (API.STATUS_FAIL, message, None)
        
        return (API.STATUS_SUCCESS, "Success", json_string)
        
    def instance_start(self, inst_id, extra_files, run_cmds):
        
        (success, message, inst) = self.__get_instance(inst_id)
        
        if not success:
            return (API.STATUS_FAIL, message)

        log.set_logging_instance(inst)
            
        try:
            deployer_class = self.__get_deployer_class(inst)
            deployer = deployer_class(extra_files, run_cmds)
            
            try:
                deployer.set_instance(inst)
            except DeploymentException, de:
                message = "Deployer failed to initialize. %s " % de
                return (API.STATUS_FAIL, message)               
            
            if inst.topology.state == Topology.STATE_NEW:
                resuming = False
            elif inst.topology.state == Topology.STATE_STOPPED:
                resuming = True
            else:
                message = "Cannot start an instance that is in state '%s'" % (Topology.state_str[inst.topology.state])
                return (API.STATUS_FAIL, message)
            
            if not resuming:
                inst.topology.state = Topology.STATE_STARTING
            else:
                inst.topology.state = Topology.STATE_RESUMING
    
            inst.topology.save()

            if not resuming:
                try:
                    eps = inst.topology.get_go_endpoints()        
                    self.__globusonline_pre_start(inst, eps)
                except GlobusOnlineException, goe:
                    log.warning("Unable to create GO endpoint/s: %s" % goe)
               
            inst.topology.save()                            
            
            nodes = inst.topology.get_nodes()
    
            (success, message, node_vm) = self.__allocate_vms(deployer, nodes, resuming)
            
            if not success:
                inst.topology.state = Topology.STATE_FAILED
                inst.topology.save()
                return (API.STATUS_FAIL, message)
    
            inst.topology.state = Topology.STATE_CONFIGURING
            inst.topology.save()
                          
            log.info("Instances are running.")
    
            for node, vm in node_vm.items():
                deployer.post_allocate(node, vm)

            inst.topology.save()
                            
            # Generate certificates
            if not resuming:
                inst.gen_certificates(force_hosts=False, force_users=False)
            else:
                inst.gen_certificates(force_hosts=True, force_users=False)    
            
            inst.topology.gen_chef_ruby_file(inst.instance_dir + "/topology.rb")
            inst.topology.gen_hosts_file(inst.instance_dir + "/hosts")               
    
            log.info("Setting up Globus Provision on instances")        
            
            (success, message) = self.__configure_vms(deployer, node_vm)
            if not success:
                inst.topology.state = Topology.STATE_FAILED
                inst.topology.save()
                return (API.STATUS_FAIL, message)
    
            inst.topology.state = Topology.STATE_RUNNING
            inst.topology.save()         
            
            log.info("Creating Globus Online endpoints")
            eps = inst.topology.get_go_endpoints()        
            if not resuming:
                try:
                    self.__globusonline_post_start(inst, eps)
                except GlobusOnlineException, goe:
                    log.warning("Unable to create GO endpoint/s: %s" % goe)
            else:        
                try:
                    self.__globusonline_resume(inst, eps)
                except GlobusOnlineException, goe:
                    log.warning("Unable to resume GO endpoint/s: %s" % goe)            

            inst.topology.save()       
            
            return (API.STATUS_SUCCESS, "Success")
        except:
            message = self.__unexpected_exception_to_text(what = "starting the instance.")
            try:
                if inst != None:
                    inst.topology.state = Topology.STATE_FAILED
                    inst.topology.save()
            except:
                message += "\nNote: Unable to update instance's state to 'Failed'"
            return (API.STATUS_FAIL, message)


    def instance_update(self, inst_id, topology_json, extra_files, run_cmds):
        try:
            (success, message, inst) = self.__get_instance(inst_id)
            
            if not success:
                return (API.STATUS_FAIL, message)
    
            log.set_logging_instance(inst)
                            
            if inst.topology.state == Topology.STATE_NEW:
                # If the topology is still in a New state, we simply
                # validate that the update is valid, and replace
                # the old topology. We don't need to deploy or
                # configure any hosts..
                if topology_json != None:
                    (success, message, create_hosts, destroy_hosts) = inst.update_topology(topology_json)
                    if not success:
                        message = "Error in topology file: %s" % message
                        return (API.STATUS_FAIL, message)
                        
                return (API.STATUS_SUCCESS, "Success")
            elif inst.topology.state != Topology.STATE_RUNNING:
                message = "Cannot update the topology of an instance that is in state '%s'" % (Topology.state_str[inst.topology.state])
                return (API.STATUS_FAIL, message)        
    
            deployer_class = self.__get_deployer_class(inst)
            deployer = deployer_class(extra_files, run_cmds)
            
            try:
                deployer.set_instance(inst)
            except DeploymentException, de:
                message = "Deployer failed to initialize. %s " % de
                return (API.STATUS_FAIL, message)    
                
            if topology_json != None:
                old_topology = inst.topology
                try:
                    (success, message, topology_changes) = inst.update_topology(topology_json)
                    if not success:
                        return (API.STATUS_FAIL, message)
                except ObjectValidationException, ove:
                    message = "Error in topology file: %s" % ove
                    return (API.STATUS_FAIL, message)
                
                create_hosts = []
                destroy_hosts = []
                
                create_endpoints = []
                remove_endpoints = []
    
                if topology_changes.changes.has_key("domains"):
                    for domain in topology_changes.changes["domains"].add:
                        d = inst.topology.domains[domain]
                        create_hosts += [n.id for n in d.nodes.values()] 
        
                    for domain in topology_changes.changes["domains"].remove:
                        d = inst.topology.domains[domain].keys()
                        destroy_hosts += [n.id for n in d.nodes.values()] 
                    
                    for domain in topology_changes.changes["domains"].edit:
                        if topology_changes.changes["domains"].edit[domain].changes.has_key("nodes"):
                            nodes_changes = topology_changes.changes["domains"].edit[domain].changes["nodes"]
                            create_hosts += nodes_changes.add
                            destroy_hosts += nodes_changes.remove
                            
                        if topology_changes.changes["domains"].edit[domain].changes.has_key("go_endpoints"):
                            ep_changes = topology_changes.changes["domains"].edit[domain].changes["go_endpoints"]
                            if ep_changes.change_type == PropertyChange.ADD:
                                create_endpoints += inst.topology.domains[domain].go_endpoints
                            elif ep_changes.change_type == PropertyChange.REMOVE:
                                remove_endpoints += old_topology.domains[domain].go_endpoints            
                            elif ep_changes.change_type == PropertyChange.EDIT:
                                create_endpoints += ep_changes.add
                                remove_endpoints += ep_changes.remove                            
    
                nodes = inst.topology.get_nodes()
    
                if len(destroy_hosts) > 0:
                    old_nodes = old_topology.get_nodes()
                    log.info("Terminating hosts %s" % destroy_hosts)   
                    old_nodes = [n for n in old_nodes if n.id in destroy_hosts]
                    (success, message) = self.__terminate_vms(deployer, old_nodes)
                    
                    if not success:
                        inst.topology.state = Topology.STATE_FAILED
                        inst.topology.save()
                        return (API.STATUS_FAIL, message)       
                    
                    inst.topology.save()         

                if len(create_endpoints) > 0:
                    try:      
                        self.__globusonline_pre_start(inst, create_endpoints)
                    except GlobusOnlineException, goe:
                        log.warning("Unable to create GO endpoint/s: %s" % goe)                      
                      
                if len(create_hosts) > 0:
                    nodes = inst.topology.get_nodes()
                    log.info("Allocating VMs for hosts %s" % create_hosts)   
                    new_nodes = [n for n in nodes if n.id in create_hosts]
                    (success, message, node_vm) = self.__allocate_vms(deployer, new_nodes, resuming = False)
            
                    if not success:
                        inst.topology.state = Topology.STATE_FAILED
                        inst.topology.save()
                        return (API.STATUS_FAIL, message)
                
                    inst.topology.save()
                    
                    for node, vm in node_vm.items():
                        deployer.post_allocate(node, vm)
    
                    inst.topology.save()
    
                # Generate certificates
                inst.gen_certificates()
    
                inst.topology.gen_chef_ruby_file(inst.instance_dir + "/topology.rb")
                inst.topology.gen_hosts_file(inst.instance_dir + "/hosts")               

            log.info("Setting up Globus Provision on instances")        
            
            # Right now we reconfigure all nodes. It shouldn't be hard to follow
            # the dependency tree to make sure only the new nodes and "ancestor"
            # nodes are updated
            nodes = inst.topology.get_nodes()
            node_vm = deployer.get_node_vm(nodes)
            (success, message) = self.__configure_vms(deployer, node_vm)
            if not success:
                inst.topology.state = Topology.STATE_FAILED
                inst.topology.save()
                return (API.STATUS_FAIL, message)
    
            if topology_json != None:
                self.__globusonline_remove(inst, remove_endpoints)
    
                if len(create_endpoints) > 0:
                    try:      
                        self.__globusonline_post_start(inst, create_endpoints)
                    except GlobusOnlineException, goe:
                        log.warning("Unable to create GO endpoint/s: %s" % goe)    
    
            inst.topology.state = Topology.STATE_RUNNING
            inst.topology.save()
            
            return (API.STATUS_SUCCESS, "Success")
        except:
            message = self.__unexpected_exception_to_text(what = "starting the instance.")
            try:
                if inst != None:
                    inst.topology.state = Topology.STATE_FAILED
                    inst.topology.save()
            except:
                message += "\nNote: Unable to update instance's state to 'Failed'"
            return (API.STATUS_FAIL, message)


    def instance_stop(self, inst_id):
        (success, message, inst) = self.__get_instance(inst_id)
        
        if not success:
            return (API.STATUS_FAIL, message)

        log.set_logging_instance(inst)
        
        try:
            if inst.topology.state != Topology.STATE_RUNNING:
                message = "Cannot start an instance that is in state '%s'" % (Topology.state_str[inst.topology.state])
                return (API.STATUS_FAIL, message)

            deployer_class = self.__get_deployer_class(inst)
            deployer = deployer_class()
            
            try:
                deployer.set_instance(inst)
            except DeploymentException, de:
                message = "Deployer failed to initialize. %s " % de
                return (API.STATUS_FAIL, message)       
        
            inst.topology.state = Topology.STATE_STOPPING
            inst.topology.save()
            
            nodes = inst.topology.get_nodes()            
    
            (success, message) = self.__stop_vms(deployer, nodes)
            
            if not success:
                inst.topology.state = Topology.STATE_FAILED
                inst.topology.save()
                return (API.STATUS_FAIL, message)            
        
            inst.topology.state = Topology.STATE_STOPPED
            inst.topology.save()
              
            log.info("Stopping Globus Online endpoints")
            try:
                eps = inst.topology.get_go_endpoints()        
                self.__globusonline_stop(inst, eps)     
                inst.topology.save()       
            except GlobusOnlineException, goe:
                log.warning("Unable to stop GO endpoint/s: %s" % goe)              
              
            log.info("Instance has stopped successfully.")
            return (API.STATUS_SUCCESS, "Success")
        except:
            message = self.__unexpected_exception_to_text(what = "stopping the instance.")
            try:
                if inst != None:
                    inst.topology.state = Topology.STATE_FAILED
                    inst.topology.save()
            except:
                message += "\nNote: Unable to update instance's state to 'Failed'"
            return (API.STATUS_FAIL, message)


    def instance_terminate(self, inst_id):
        (success, message, inst) = self.__get_instance(inst_id)
        
        if not success:
            return (API.STATUS_FAIL, message)

        log.set_logging_instance(inst)
        
        try:
            if inst.topology.state in [Topology.STATE_NEW]:
                message = "Cannot terminate an instance that is in state '%s'" % (Topology.state_str[inst.topology.state])
                return (API.STATUS_FAIL, message)

            deployer_class = self.__get_deployer_class(inst)
            deployer = deployer_class()
            
            try:
                deployer.set_instance(inst)
            except DeploymentException, de:
                message = "Deployer failed to initialize. %s " % de
                return (API.STATUS_FAIL, message)       
        
            inst.topology.state = Topology.STATE_TERMINATING
            inst.topology.save()
            
            nodes = inst.topology.get_nodes()            
    
            (success, message) = self.__terminate_vms(deployer, nodes)
            
            if not success:
                inst.topology.state = Topology.STATE_FAILED
                inst.topology.save()
                return (API.STATUS_FAIL, message)            
        
            # Remove GO endpoints
            eps = inst.topology.get_go_endpoints()        
            self.__globusonline_remove(inst, eps)
        
            inst.topology.state = Topology.STATE_TERMINATED
            inst.topology.save()
              
            log.info("Instances have been terminated.")
            return (API.STATUS_SUCCESS, "Success")
        except:
            message = self.__unexpected_exception_to_text(what = "starting the instance.")
            try:
                if inst != None:
                    inst.topology.state = Topology.STATE_FAILED
                    inst.topology.save()
            except:
                message += "\nNote: Unable to update instance's state to 'Failed'"
            return (API.STATUS_FAIL, message)


    def instance_list(self, inst_ids):
        istore = InstanceStore(self.instances_dir)
        
        (valid, invalid) = istore.get_instances(inst_ids)
        instances_jsons = []
        for inst in valid:
            instances_jsons.append(inst.topology.to_json_string())
            
        instances_json = "[" + ",".join(instances_jsons) + "]"
                
        # TODO: Return invalid instances as "warnings", once we
        # add that extra return value to the API.
        return API.STATUS_SUCCESS, "Success", instances_json

        
    def __get_instance(self, inst_id):
        try:
            istore = InstanceStore(self.instances_dir)
            inst = istore.get_instance(inst_id)
            return (True, "Success", inst)
        except ConfigException, cfge:
            message = "Error in configuration file: %s" % cfge
            return (False, message, None)
        except ObjectValidationException, ove:
            message = "Error in topology file: %s" % ove
            return (False, message, None)        
        except InstanceException, ie:
            return (False, str(ie), None)
        except:
            message = self.__unexpected_exception_to_text(what = "accessing the instance.")
            return (False, message, None)    
        
    def __get_deployer_class(self, inst):
        if inst.config.get("deploy") == "ec2":
            deploy_module = ec2_deploy
        elif inst.config.get("deploy") == "dummy":
            deploy_module = dummy_deploy
            
        return deploy_module.Deployer        
        

    def __globusonline_pre_start(self, inst, eps):
        go_helper = GlobusOnlineHelper.from_instance(inst)

        for ep in eps:        
            if ep.has_property("globus_connect_cert") and ep.globus_connect_cert:
                if ep.gridftp.startswith("node:"):
                    gridftp_node = inst.topology.get_node_by_id(ep.gridftp[5:])
                    go_helper.connect(ep.user)
                    gc_setupkey = go_helper.endpoint_gc_create(ep, replace = True)
                    gridftp_node.set_property("gc_setupkey", gc_setupkey)
                    go_helper.disconnect()        


    def __globusonline_post_start(self, inst, eps):
        go_helper = GlobusOnlineHelper.from_instance(inst)
        
        for ep in eps:
            go_helper.connect(ep.user)
            if ep.has_property("globus_connect_cert") and ep.globus_connect_cert:
                if ep.gridftp.startswith("node:"):
                    go_helper.endpoint_gc_create_finalize(ep)
            else:        
                go_helper.create_endpoint(ep, replace=True)
            go_helper.disconnect()
                    
    def __globusonline_remove(self, inst, eps):
        go_helper = GlobusOnlineHelper.from_instance(inst)
        
        try:
            for ep in eps:     
                go_helper.connect(ep.user)
                try:
                    go_helper.endpoint_remove(ep)
                except:
                    pass   
                go_helper.disconnect()
        except GlobusOnlineException, goe:
            log.warning("Unable to remove GO endpoint/s: %s" % goe)        
        
                    
    def __globusonline_stop(self, inst, eps):
        go_helper = GlobusOnlineHelper.from_instance(inst)
        
        for ep in eps:
            go_helper.connect(ep.user)
            go_helper.endpoint_stop(ep)
            go_helper.disconnect()                    

    def __globusonline_resume(self, inst, eps):
        go_helper = GlobusOnlineHelper.from_instance(inst)
        
        for ep in eps:
            go_helper.connect(ep.user)
            go_helper.endpoint_resume(ep)
            go_helper.disconnect()                 

        
    def __allocate_vms(self, deployer, nodes, resuming):
        # TODO: Make this an option
        sequential = False
        topology = deployer.instance.topology
        
        if not resuming:
            log.info("Allocating %i VMs." % len(nodes))
            next_state = Node.STATE_RUNNING_UNCONFIGURED
        else:
            log.info("Resuming %i VMs" % len(nodes))
            next_state = Node.STATE_RESUMED_UNCONFIGURED
        node_vm = {}
        for n in nodes:
            try:
                if not resuming:
                    n.set_property("state", Node.STATE_STARTING)
                    topology.save()
                    vm = deployer.allocate_vm(n)
                else:
                    n.set_property("state", Node.STATE_RESUMING)
                    topology.save()
                    vm = deployer.resume_vm(n)
                node_vm[n] = vm
            except Exception:
                message = self.__unexpected_exception_to_text()
                return (False, message, None)
        
            if sequential:
                log.debug("Waiting for instance to start.")
                wait = deployer.NodeWaitThread(None, "wait-%s" % str(vm), n, vm, deployer, state = next_state)
                wait.run2()
                
        if not sequential:        
            log.debug("Waiting for instances to start.")
            mt_instancewait = MultiThread()
            for node, vm in node_vm.items():
                mt_instancewait.add_thread(deployer.NodeWaitThread(mt_instancewait, "wait-%s" % str(vm), node, vm, deployer, state = next_state))

            mt_instancewait.run()
            if not mt_instancewait.all_success():
                message = self.__mt_exceptions_to_text(mt_instancewait.get_exceptions(), "Exception raised while waiting for instances.")
                return (False, message, None)
            
        return (True, "Success", node_vm)


    def __configure_vms(self, deployer, node_vm, basic = True, chef = True):
        nodes = node_vm.keys()
        mt_configure = MultiThread()
        topology = deployer.instance.topology
        order = topology.get_launch_order(nodes)
        
        threads = {}
        for nodeset in order:
            rest = dict([(n, deployer.NodeConfigureThread(mt_configure, 
                            "configure-%s" % n.id, 
                            n, 
                            node_vm[n], 
                            deployer, 
                            depends=threads.get(topology.get_depends(n)),
                            basic = basic,
                            chef = chef)) for n in nodeset])
            threads.update(rest)
        
        for thread in threads.values():
            mt_configure.add_thread(thread)
        
        mt_configure.run()
        if not mt_configure.all_success():
            message = self.__mt_exceptions_to_text(mt_configure.get_exceptions(), "Globus Provision was unable to configure the instances.")
            return (False, message)
        
        return (True, "Success")

    def __mt_exceptions_to_text(self, exceptions, what):
        msg = "ERROR - " + what
        for thread_name in exceptions:
            msg += "\n\n"
            exception_obj, exception_trace = exceptions[thread_name]
            if isinstance(exception_obj, SSHCommandFailureException):
                msg += "        %s: Error while running '%s'\n" % (thread_name, exception_obj.command)
            elif isinstance(exception_obj, EC2ResponseError):
                msg += "        %s: EC2 error '%s'\n" % (thread_name, exception_obj.reason)
                msg += "        Body: %s\n" % exception_obj.body
            else:          
                msg += "        %s: Unexpected exception '%s'\n" % (thread_name, exception_obj.__class__.__name__)
                msg += "        Message: %s\n" % exception_obj
            for l in exception_trace:
                msg += l
        return msg

            
    def __unexpected_exception_to_text(self, what=""):
        exc_type, exc_value, exc_traceback = sys.exc_info()
        if what != "": what = " when " + what
        msg = "An unexpected '%s' exception has been raised%s\n\n" % (exc_type.__name__, what)
        msg += "Message: %s\n\n" % exc_value
        trace = traceback.format_exception(exc_type, exc_value, exc_traceback)
        for l in trace:
            msg += l
        return msg


    def __stop_vms(self, deployer, nodes):
        node_vm = deployer.get_node_vm(nodes)
        topology = deployer.instance.topology
        mt_configure = MultiThread()        
        order = topology.get_launch_order(nodes)
        order.reverse()
        
        for n in node_vm:
            n.state = Node.STATE_STOPPING
        topology.save()
        
        threads = {}
        for nodeset in order:
            rest = dict([(n, deployer.NodeConfigureThread(mt_configure, 
                            "stop-configure-%s" % n.id, 
                            n, 
                            node_vm[n], 
                            deployer, 
                            depends=threads.get(topology.get_depends(n)))) for n in nodeset])
            threads.update(rest)
        
        for thread in threads.values():
            mt_configure.add_thread(thread)
        
        mt_configure.run()
        if not mt_configure.all_success():
            message = self.__mt_exceptions_to_text(mt_configure.get_exceptions(), "Globus Provision was unable to configure the instances.")
            return (False, message)        
        
        
        for nodeset in order:
            deployer.stop_vms(nodeset)
        
        log.debug("Waiting for instances to stop.")
        mt_instancewait = MultiThread()
        for node, vm in node_vm.items():
            mt_instancewait.add_thread(deployer.NodeWaitThread(mt_instancewait, "wait-%s" % str(vm), node, vm, deployer, state = Node.STATE_STOPPED))
        
        mt_instancewait.run()
        if not mt_instancewait.all_success():
            message = self.__mt_exceptions_to_text(mt_instancewait.get_exceptions(), "Exception raised while waiting for instances.")
            return (False, message)     
            
        return (True, "Success")
        
    def __terminate_vms(self, deployer, nodes):
        topology = deployer.instance.topology

        for n in nodes:
            n.state = Node.STATE_TERMINATING
        topology.save()

        deployer.terminate_vms(nodes)
        
        node_vm = deployer.get_node_vm(nodes)
        
        log.debug("Waiting for instances to terminate.")
        mt_instancewait = MultiThread()
        for node, vm in node_vm.items():
            mt_instancewait.add_thread(deployer.NodeWaitThread(mt_instancewait, "wait-%s" % str(vm), node, vm, deployer, state = Node.STATE_TERMINATED))
        
        mt_instancewait.run()
        if not mt_instancewait.all_success():
            message = self.__mt_exceptions_to_text(mt_instancewait.get_exceptions(), "Exception raised while waiting for instances.")
            return (False, message)
            
        return (True, "Success")
        
        


from globus.provision import AMI
import os

ami = AMI["us-east-1"]["32-bit"]

sample_files = os.listdir("../samples/")

for fname in sample_files:
    fname1 = "../samples/" + fname
    fname2 = "./_build/samples/" + fname

    f1 = open(fname1, "r")
    f2 = open(fname2, "w")
    
    for l in f1:
        f2.write(l.replace("%AMI%", ami))
    
    f1.close()
    f2.close()



# -------------------------------------------------------------------------- #
# Copyright 2010-2011, University of Chicago                                 #
#                                                                            #
# Licensed under the Apache License, Version 2.0 (the "License"); you may    #
# not use this file except in compliance with the License. You may obtain    #
# a copy of the License at                                                   #
#                                                                            #
# http://www.apache.org/licenses/LICENSE-2.0                                 #
#                                                                            #
# Unless required by applicable law or agreed to in writing, software        #
# distributed under the License is distributed on an "AS IS" BASIS,          #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.   #
# See the License for the specific language governing permissions and        #
# limitations under the License.                                             #
# -------------------------------------------------------------------------- #
"""
Commands related to Globus Online endpoint management, but which do not require access to the API
"""

import sys
import os.path
import paramiko
from globus.provision.common.ssh import SSH

from globus.provision.cli import Command
from globus.provision.core.instance import InstanceStore
from globus.provision.common.go_transfer import GlobusOnlineHelper

def gp_go_register_endpoints_func():
    return gp_go_register_endpoints(sys.argv).run()     

class gp_go_register_endpoints(Command):
    """
    Creates the Globus Online endpoints specified in an instance's topology.
    
    The instance identifier must be specified after all other parameters. For example::
    
        gp-go-register-endpoints --public gpi-12345678    
    """     
    
    name = "gp-go-register-endpoints"
    
    def __init__(self, argv):
        Command.__init__(self, argv)    

        self.optparser.add_option("-m", "--domain", 
                                  action="store", type="string", dest="domain", 
                                  help = "Register only the endpoints in this domain") 

        self.optparser.add_option("-r", "--replace", 
                                  action="store_true", dest="replace", 
                                  help = "If an endpoint already exists, replace it")           
                
    def run(self):    
        self.parse_options()
        inst_id = self.args[1]
        
        istore = InstanceStore(self.opt.dir)
        inst = istore.get_instance(inst_id)        
        
        go_helper = GlobusOnlineHelper.from_instance(inst)

        for domain_name, domain in inst.topology.domains.items():
            for ep in domain.go_endpoints:
                go_helper.connect(ep.user)
                if (not ep.has_property("globus_connect_cert")) or (ep.has_property("globus_connect_cert") and not ep.globus_connect_cert):
                    go_helper.create_endpoint(ep, self.opt.replace)
                go_helper.disconnect()
                        
                print "Created endpoint '%s#%s' for domain '%s'" % (ep.user, ep.name, domain_name)


# -------------------------------------------------------------------------- #
# Copyright 2010-2011, University of Chicago                                 #
#                                                                            #
# Licensed under the Apache License, Version 2.0 (the "License"); you may    #
# not use this file except in compliance with the License. You may obtain    #
# a copy of the License at                                                   #
#                                                                            #
# http://www.apache.org/licenses/LICENSE-2.0                                 #
#                                                                            #
# Unless required by applicable law or agreed to in writing, software        #
# distributed under the License is distributed on an "AS IS" BASIS,          #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.   #
# See the License for the specific language governing permissions and        #
# limitations under the License.                                             #
# -------------------------------------------------------------------------- #

"""
Miscellaneous utility functions.
"""

import glob
import os

from os import environ

from boto.ec2.connection import EC2Connection,RegionInfo
from boto import connect_ec2
       

def create_ec2_connection(hostname = None, path = None, port = None):
    if hostname == None:
        # We're using EC2.
        # Check for AWS_ACCESS_KEY_ID and AWS_SECRET_ACCESS_KEY,
        # and use EC2Connection. boto will fill in all the values
        if not (environ.has_key("AWS_ACCESS_KEY_ID") and environ.has_key("AWS_SECRET_ACCESS_KEY")):
            return None
        else:
            return EC2Connection()
    else:
        # We're using an EC2-ish cloud.
        # Check for EC2_ACCESS_KEY and EC2_SECRET_KEY (these are used by Eucalyptus;
        # we will probably have to tweak this further to support other systems)
        if not (environ.has_key("EC2_ACCESS_KEY") and environ.has_key("EC2_SECRET_KEY")):
            return None
        else:
            print "Setting region"
            region = RegionInfo(name="eucalyptus", endpoint=hostname)
            return connect_ec2(aws_access_key_id=environ["EC2_ACCESS_KEY"],
                        aws_secret_access_key=environ["EC2_SECRET_KEY"],
                        is_secure=False,
                        region=region,
                        port=port,
                        path=path)            


def parse_extra_files_files(f):
    l = []
    extra_f = open(f)
    for line in extra_f:
        srcglob, dst = line.split()
        srcs = glob.glob(os.path.expanduser(srcglob))
        srcs = [s for s in srcs if os.path.isfile(s)]
        dst_isdir = (os.path.basename(dst) == "")
        for src in srcs:
            full_dst = dst
            if dst_isdir:
                full_dst += os.path.basename(src)
            l.append( (src, full_dst) )
    return l

def rest_table(col_names, rows):
    def gen_line(lens, char):
        return "+" + char + (char + "+" + char).join([char * l for l in lens]) + char + "+\n"
    
    num_cols = len(col_names)
    len_cols = [0] * num_cols
    height_row = [0] * len(rows)
    
    table = ""
    
    for i, name in enumerate(col_names):
        len_cols[i] = max(len(name), len_cols[i])
        
    for i, row in enumerate(rows):
        for j in range(num_cols):
            lines = row[j].split("\n")
            row_len = max([len(l) for l in lines])
            len_cols[j] = max(row_len, len_cols[j])
            height_row[i] = max(len(lines), height_row[i])
            
    table += gen_line(len_cols, "-")

    table += "|"
    for i, name in enumerate(col_names): 
        table += " "
        table += col_names[i].ljust(len_cols[i])
        table += " |"
    table += "\n"

    table += gen_line(len_cols, "=")
    
    for i, row in enumerate(rows):
        for j in range(height_row[i]):
            table += "|"
            for k, col in enumerate(row): 
                lines = col.split("\n")
                
                if len(lines) < j + 1:
                    col_txt = " " * len_cols[k]
                else:
                    col_txt = lines[j].ljust(len_cols[k])
                
                table += " "
                table += col_txt
                table += " |"
            table += "\n"            

        table += gen_line(len_cols, "-")
    
    return table

# From http://stackoverflow.com/questions/36932/whats-the-best-way-to-implement-an-enum-in-python
def enum(*sequential, **named):
    enums = dict(zip(sequential, range(len(sequential))), **named)
    return type('Enum', (), enums)





# -*- encoding: utf-8
# Copyright (C) 2009-2011  Rene Koecher <shirk@bitspin.org>
#
# This file is part of paraproxy.
#
# Paraproxy is free software; you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 2.1 of the License, or (at your option)
# any later version.
#
# Paraproxy is distrubuted in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with paraproxy; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.

import paramiko

import re
import os
import time
import socket
import signal
import thread
import fnmatch
import getpass
import subprocess
import shlex
import StringIO

from WrapSock import WrapSock

# for  debugging
__paramiko_SSHClient_connect_verbose = os.getenv('PARAMIKO_SSH_HOOK_VERBOSE', False)

# hook for original SSHClient.connect function
__paramiko_SSHClient_connect_hook = globals()['paramiko'].SSHClient.connect
# hook for original SSHClient.close function
__paramiko_SSHClient_close_hook = globals()['paramiko'].SSHClient.close
# hook for original SSHClient.__del__ function (if any)
try:
	__paramkio_SSHClient_delete_hook = globals()['paramiko'].SSHClient.__del__
except AttributeError:
	__paramiko_SSHClient_delete_hook = None

# list with configured proxy-commands / ssh option etc.
__paramiko_SSHClient_proxy_config = paramiko.SSHConfig()

# Import paramiko internals with prefix (__.*)
from paramiko.hostkeys import HostKeys as __HostKeys
from paramiko.resource import ResourceManager as __ResourceManager
from paramiko.transport import Transport as __Transport
from paramiko.ssh_exception import SSHException, BadHostKeyException

def __paramiko_SSHClient_debug(msg):
	if __paramiko_SSHClient_connect_verbose:
		print '[%s] %s' % (os.getpid(), msg)

def __paramiko_SSHClient_proxy_cmdchain(self, hostname, rec=0):
	"""
	Resolver used to detect multi-layer ProxyCommand chains

	This will search for hostname in proxy_list, if found
	add the assigned ProxyCommand to the chain and check if it it turn
	requires a ProxyCommand

	Chain detection depends on the command to be 'ssh'/'ssh.exe' (absolute paths are supported)
	"""
	chain = []
	command = __paramiko_SSHClient_proxy_config.lookup(hostname).get('proxycommand')
	port = __paramiko_SSHClient_proxy_config.lookup(hostname).get('port', '22')
	hostname = __paramiko_SSHClient_proxy_config.lookup(hostname).get('hostname', hostname)
	if not command:
		if rec != 0:
			__paramiko_SSHClient_debug('__paramiko_SSHClient_proxy_cmdchain(): hop_%d: %s:%s, DIRECT' % (rec,hostname,port))
			return [(hostname, 'DIRECT', None)]
		return []

	# command found
	try:
		#__paramiko_SSHClient_debug('processing "%s"' % command)
		command_args = [] + shlex.split(command)

		exe = os.path.basename(command_args[0].lower())
		if exe.lower() == 'ssh' or exe.lower() == 'ssh.exe':
			# command is SSH, extract the host and the remote command
			remote_cmd = command_args.pop()
			remote_host = command_args.pop()
			if '@' in remote_host:
			    proxyuser, remote_host = remote_host.split("@")
			else:
			    proxyuser = None

			chain = [(remote_host, remote_cmd.replace('%h', hostname).replace('%p', port), proxyuser)]
			next_hop = __paramiko_SSHClient_proxy_cmdchain(self, remote_host, rec+1)

			if next_hop:
				chain = next_hop + chain
		else:
			# a regular command like netcat
			chain = [(hostname, command.replace('%h', hostname).replace('%p', port), None)]
			if rec != 0:
			    __paramiko_SSHClient_debug('__paramiko_SSHClient_proxy_cmdchain(): WARNING: regular command "%s" for "%s" will be run in hop_%d'  % (command, hostname, rec))
			    __paramiko_SSHClient_debug('__paramiko_SSHClient_proxy_cmdchain(): WARNING: paraproxy doesn\'t know how to chain further from here.')
			return chain
	except Exception,e:
		#FIXME: propper handling!
		 __paramiko_SSHClient_debug('__paramiko_SSHClient_proxy_cmdchain(): ERROR: failed to parse ProxyCommand for host "%s" - abort!' % (hostname))
		 raise

	if rec != 0:
		__paramiko_SSHClient_debug('__paramiko_SSHClient_proxy_cmdchain(): hop_%d: %s:%s' % (rec,hostname, port))
	else:
		__paramiko_SSHClient_debug('__paramiko_SSHClient_proxy_cmdchain(): target: %s:%s' % (hostname, port))
		__paramiko_SSHClient_debug('__paramiko_SSHClient_proxy_cmdchain(): final chain: %s' % (' => '.join([('(%s, %s)' % (x[0],x[1])) for x in chain])))

	return chain

def __paramiko_SSHClient_proxy_thread(self, ip, port, command):
	"""
	Proxy thread used to establish a connection using a ssh ProxyCommand.
	
	This will open a unix-domain-socket and listen for one incomming connection.
	After this it will spawn the proxy command an wrap the open socket to act
	as stdin and stdout for the new process.
	"""
	__paramiko_SSHClient_debug('proxy_thread(%s,%s)' % (ip, port))
	self._sock_name = os.path.join(os.path.sep,'tmp', '.paramiko_proxy_%s_%s_%s' % (ip, port, os.getpid()))
	self._should_terminate = False

	try:
		os.unlink(self._sock_name)
	except OSError:
		pass
	
	sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
	sock.bind(self._sock_name)
	sock.listen(1)
	__paramiko_SSHClient_debug('proxy: bound to `%s` and listening' % self._sock_name)
	
	(stmp, addr) = sock.accept()
	__paramiko_SSHClient_debug('proxy: incomming connection')
	
	cmd_str = str(command) # copy
	self._proc_id = None
	wrap_sock = WrapSock(stmp)

	try:
		__paramiko_SSHClient_debug('proxy: executing command "%s"' % cmd_str.replace('%h', str(ip)).replace('%p', str(port)))

		proc = subprocess.Popen(cmd_str.replace('%h', str(ip)).replace('%p', str(port)),
								shell=True,
								stdin=wrap_sock,
								stdout=wrap_sock,
								close_fds=True)
	
		self._proc_id = proc.pid
		self._sock_state = 1

		# warten auf Beendigung der Verbindung
		while proc.poll() == None \
		      and not self._should_terminate:
			time.sleep(0.5)

		if self._should_terminate:
			__paramiko_SSHClient_debug('proxy: termination requested..')
		else:
			__paramiko_SSHClient_debug('proxy: ProxyCommand terminated, exit code: %s' % proc.returncode)

		del proc

	except OSError, e:
		__paramiko_SSHClient_debug('proxy: OSError in Popen: %s' % e)
	
	except ValueError, e:
		__paramiko_SSHClient_debug('proxy: ValueError in Popen: %s' % e)

	wrap_sock.close()
	sock.close()
	
	try:
		os.unlink(self._sock_name)
	except OSError,e:
		__paramiko_SSHClient_debug('proxy: error unlinking %s: %s' % (self._sock_name, str(e)))
	
	self._proc_id = None
	__paramiko_SSHClient_debug('proxy: ~proxy_thread()')


def __paramiko_SSHClient_connect(self, hostname, port=22, username=None,
				 password=None, pkey=None, key_filename=None,
				 timeout=None, allow_agent=True, look_for_keys=True):
	"""
	Hook function to enable Paramiko's SSHClient to work with ssh ProxyCommand.
	This will only hook into connections matching one of the ProxyRules.
	"""

	def __create_and_connect_transport(self, _hostname, _transport_sock, _username=None, _password=None):
		"""
		Helper function to create a new Transport and perform the required SSH authentication
		-- just here to keep the code more readable
		"""
		#print 'Create transport on: %s' % _transport_sock
		t = self._transport = __Transport(_transport_sock)
		self._chained_transports.append(t)

		if self._log_channel is not None:
			t.set_log_channel(self._log_channel)

		t.start_client()
		__ResourceManager.register(self, t)
		
		server_key = t.get_remote_server_key()
		keytype = server_key.get_name()
		
		our_server_key = self._system_host_keys.get(_hostname, {}).get(keytype, None)
		if our_server_key is None:
			our_server_key = self._host_keys.get(_hostname, {}).get(keytype, None)
		if our_server_key is None:
			# will raise exception if the key is rejected; let that fall out
			self._policy.missing_host_key(self, _hostname, server_key)
			# if the callback returns, assume the key is ok
			our_server_key = server_key
	
		if server_key != our_server_key:
			raise BadHostKeyException(_hostname, server_key, our_server_key)
	
		if _username is None:
			_username = getpass.getuser()
	
		if key_filename is None:
			key_filenames = []
		elif isinstance(key_filename, (str, unicode)):
			key_filenames = [ key_filename ]
		else:
			key_filenames = key_filename

		self._auth(_username, _password, pkey, key_filenames, allow_agent, look_for_keys)
	#---

	__paramiko_SSHClient_debug('__paramiko_SSHClient_connect(%s, %s)' % (hostname, port))
	command = __paramiko_SSHClient_proxy_config.lookup(hostname).get('proxycommand')
	
	if not command:
		__paramiko_SSHClient_debug('__paramiko_SSHClient_connect(): direct connection.')
		return __paramiko_SSHClient_connect_hook(self, hostname, port, username, password, pkey,
							 key_filename, timeout, allow_agent, look_for_keys)

	command_chain = __paramiko_SSHClient_proxy_cmdchain(self, hostname)

	if len(command_chain) > 1:
		__paramiko_SSHClient_debug('__paramiko_SSHClient_connect(): using multi-layer chain with %d hops' % (len(command_chain) - 1))
	else:
		__paramiko_SSHClient_debug('__paramiko_SSHClient_connect(): using ProxyCommand `%s`' % command)

	self._sock_name = ''
	self._sock_state = 0
	self._proc_id = 0
	self._should_terminate = False
	self._chained_transports = []

	if not command_chain:
		thread.start_new(__paramiko_SSHClient_proxy_thread, (self, hostname, port, command))
		time.sleep(1)
		
		sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
		if timeout is not None:
			try:
				sock.settimeout(timeout)
			except:
				pass

		try:
			sock.connect(self._sock_name) # set by proxy_thread
			__paramiko_SSHClient_debug('__paramiko_SSHClient_connect(): connected to proxy thread on `%s`' % self._sock_name)
			while self._sock_state == 0:
				time.sleep(0.5)

		except Exception, e:
			# except *anything* to be able to remove the proxy thread.
			__paramiko_SSHClient_debug('__paramiko_SSHClient_connect(): caught %s - closing threads.' % e.__class__.__name__)

			try:
				if (self._proc_id):
					__paramiko_SSHClient_debug('__paramiko_SSHClient_connect(): requesting thread shutdown')
					self._should_terminate = True
				else:
					__paramiko_SSHClient_debug('__paramiko_SSHClient_connect(): os._proc_id <= 0, nothing to do.')
			except OSError:
				pass

			raise

	else:
		(proxy_host, proxy_command, proxy_username) = command_chain.pop(0)
		sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
		if timeout is not None:
			try:
				sock.settimeout(timeout)
			except:
				pass

		# no try: except: here - just raise any socke errors
		proxy_port = long(__paramiko_SSHClient_proxy_config.lookup(proxy_host).get('port', '22'))
		sock.connect((proxy_host, proxy_port))
		__paramiko_SSHClient_debug('__paramiko_SSHClient_connect(): connected to `%s:%s`' % (proxy_host, proxy_port))

	try:
		transport_sock = sock
		while command_chain:
			if command_chain[0][2] == None:
				proxy_user = username
			else:
				proxy_user = command_chain[0][2]
			__create_and_connect_transport(self, command_chain[0][0], transport_sock, proxy_user, password)
			# this hop is now valid, execute the ProxyCommand
			(proxy_host, proxy_command, proxy_username) = command_chain.pop(0)

			__paramiko_SSHClient_debug('__paramiko_SSHClient_connect(): connected to %s, sending command: %s' % (proxy_host, proxy_command))
			transport_sock = self._transport.open_session()
			__paramiko_SSHClient_debug('__paramiko_SSHClient_connect(): next transport_sock: %s' % transport_sock)
			transport_sock.exec_command(proxy_command)

		# final connection, always executed
		__create_and_connect_transport(self, hostname, transport_sock, username, password)
	except Exception, e:
		# except *anything* to be able to remove the proxy thread.
		__paramiko_SSHClient_debug('__paramiko_SSHClient_connect(): caught %s - closing threads.' % e.__class__.__name__)

		try:
			if (self._proc_id):
				__paramiko_SSHClient_debug('__paramiko_SSHClient_connect(): requesting thread shutdown')
				self._should_terminate = True
			else:
				__paramiko_SSHClient_debug('__paramiko_SSHClient_connect(): os._proc_id <= 0, nothing to do.')
		except OSError:
			pass

		raise

	__paramiko_SSHClient_debug('~__paramiko_SSHClient_connect()')

def __paramiko_SSHClient_close(self):
	__paramiko_SSHClient_debug('__paramiko_SSHClient_close(): begin closing.')
	try:
		if self._chained_transports:
			self._transport = None
			while self._chained_transports:
				__paramiko_SSHClient_debug('__paramiko_SSHClient_close(): remove hop transport..')
				t = self._chained_transports.pop(0).close()
				del t
	except AttributeError:
	    pass

	__paramiko_SSHClient_close_hook(self)
	try:
		while self._proc_id:
			self._should_terminate = True
			time.sleep(0.5)
		
	except AttributeError:
		pass
	
	try:
		if os.path.exists(self._sock_name):
			os.unlink(self._sock_name)
			__paramiko_SSHClient_debug('~__paramiko_SSHClient_close(): unlink %s successful.' % self._sock_name)
	except OSError,e:
		__paramiko_SSHClient_debug('~__paramiko_SSHClient_close(): error unlinking %s: %s' % (self._sock_name, str(e)))
	except AttributeError:
		pass

	__paramiko_SSHClient_debug('__paramiko_SSHClient_close(): finished closing.')

def __paramiko_SSHClient_delete(self):
	__paramiko_SSHClient_debug('__paramiko_SSHClient_delete(): begin destruction.')
	try:
		if self._chained_transports:
			self._transport = None
			while self._chained_transports:
				__paramiko_SSHClient_debug('__paramiko_SSHClient_delete(): remove hop transport..')
				t = self._chained_transports.pop(0).close()
				del t
	except AttributeError:
	    pass

	try:
		while self._proc_id:
			self._should_terminate = True
			time.sleep(0.5)
		
	except AttributeError:
		pass
	
	try:
		if os.path.exists(self._sock_name):
			os.unlink(self._sock_name)
			__paramiko_SSHClient_debug('~__paramiko_SSHClient_delete(): unlink %s successful.' % self._sock_name)
	except OSError,e:
		__paramiko_SSHClient_debug('__paramiko_SSHClient_delete(): error unlinking %s: %s' % (self._sock_name, str(e)))
	except AttributeError:
		pass
	
	if __paramiko_SSHClient_delete_hook:
		__paramiko_SSHClient_delete_hook(self)

	__paramiko_SSHClient_debug('~__paramiko_SSHClient_delete(): finished destruction.')

# parse ssh_config
merged_config = StringIO.StringIO()
for conf_file in [os.path.join(os.path.sep, 'etc', 'ssh', 'ssh_config'),
			       os.path.join(os.getenv('HOME','/tmp'), '.ssh', 'config')]:
	try:
		config = open(conf_file)
		merged_config.write(config.read())
		merged_config.write('\n')
		config.close()
		del config

	except IOError,e:
		pass

merged_config.seek(0,0)
__paramiko_SSHClient_proxy_config.parse(merged_config)
del merged_config

# finally modify SSHClient to use our connect hook
setattr(globals()['paramiko'].SSHClient, 'connect', __paramiko_SSHClient_connect)
setattr(globals()['paramiko'].SSHClient, 'close'  , __paramiko_SSHClient_close)
setattr(globals()['paramiko'].SSHClient, '__del__', __paramiko_SSHClient_delete)

__paramiko_SSHClient_debug('Hookhookhook')




from globus.provision.core.config import GPConfig, SimpleTopologyConfig
from globus.provision.common.config import OPTTYPE_INT, OPTTYPE_FLOAT, OPTTYPE_STRING, OPTTYPE_BOOLEAN
import textwrap
import re
import sys

conf_cls = sys.argv[1]
title = sys.argv[2]

if conf_cls == "GPConfig":
    conf_cls = GPConfig
elif conf_cls == "SimpleTopologyConfig":
    conf_cls = SimpleTopologyConfig

def print_section(title, marker):
    print title
    print marker * len(title)

for s in conf_cls.sections:
    print ".. _%s_section_%s:" % (conf_cls.__name__, s.name)
    print
    print_section("Section ``[%s]``" % s.name, "=")
    print
    print s.get_doc()
    print
    if s.required:
        print "*This section is required*"
    else:
        if s.required_if:
            print "This section is required when:"
            print
            for r in s.required_if:
                sec,opt = r[0]
                val = r[1]
                print "* Option ``%s`` (in section ``[%s]``)" % (opt,sec)
                print "  is set to ``%s``" % val
            print
        else:
            print "This section is optional."
    print
    
    for opt in s.options:
        print ".. _%s_%s:" % (conf_cls.__name__, opt.getter)
        print
        print_section("Option ``%s``" % opt.name, "-")
        
        if opt.valid:
            print "**Valid values:**",
            print "  " + ", ".join(["``%s``" % v for v in opt.valid])
        else:
            print "**Type:**",
            if opt.type == OPTTYPE_INT:
                print "Integer number"
            elif opt.type == OPTTYPE_FLOAT:
                print "Real number"
            elif opt.type == OPTTYPE_STRING:
                print "String"
            elif opt.type == OPTTYPE_BOOLEAN:
                print "Boolean (``True`` or ``False``)"
        print
        print "**Required:**",
        if opt.required:
            print "Yes"
        else:
            if opt.required_if:
                print "Only if"
                print "\\begin{itemize}"
                for r in opt.required_if:
                    sec,o = r[0]
                    val = r[1]
                    print "\\item"
                    print "Option \\texttt{%s} (in section \\texttt{[%s]})" % (o,sec)
                    print "is set to \\texttt{%s}" % val
                print "\\end{itemize}"
            else:
                txt = "No "
                if opt.default:
                    txt += "(default is ``%s``)" % opt.default
                print txt
        print
        optdoc = opt.get_doc()
        print optdoc
        print 



# -------------------------------------------------------------------------- #
# Copyright 2010-2011, University of Chicago                                 #
#                                                                            #
# Licensed under the Apache License, Version 2.0 (the "License"); you may    #
# not use this file except in compliance with the License. You may obtain    #
# a copy of the License at                                                   #
#                                                                            #
# http://www.apache.org/licenses/LICENSE-2.0                                 #
#                                                                            #
# Unless required by applicable law or agreed to in writing, software        #
# distributed under the License is distributed on an "AS IS" BASIS,          #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.   #
# See the License for the specific language governing permissions and        #
# limitations under the License.                                             #
# -------------------------------------------------------------------------- #

"""
Remote execution and file transfer via SSH

This module provides an abstraction over the paramiko package.
"""

import sys
import paramiko
import time
import select
import os.path
import traceback
from paramiko.ssh_exception import SSHException

# Try to use our patched version of paraproxy only if
# it is available. If it isn't, ProxyCommand support
# will simply be unavailable
try:
    import dg_paraproxy
except:
    pass

from Crypto.Random import atfork

from globus.provision.common import log
from os import walk

class SSHCommandFailureException(Exception):
    def __init__(self, ssh, command):
        self.ssh = ssh
        self.command = command
        
        
class SSH(object):
    def __init__(self, username, hostname, key_path, default_outf = sys.stdout, default_errf = sys.stderr, port=22):
        self.username = username
        self.hostname = hostname
        self.key_path = key_path
        self.default_outf = default_outf
        self.default_errf = default_errf
        self.port = port
        
    def open(self, timeout = 120):
        key = paramiko.RSAKey.from_private_key_file(self.key_path)
        connected = False
        t_start = time.time()
        while not connected:
            try:
                atfork()
                self.client = paramiko.SSHClient()
                self.client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
                self.client.connect(self.hostname, self.port, self.username, timeout=5, pkey=key)
                connected = True
            except Exception, e:
                t_now = time.time()
                if t_now - t_start > timeout:
                    raise e
                else:
                    time.sleep(2)

        try:
            self.sftp = self.client.get_transport().open_sftp_client()
        except SSHException, sshe:
            # Some SSH servers, like the GO CLI, are not amenable to SFTP
            log.debug("Unable to create an SFTP client on this connection.")
            self.sftp = None  
        
    def close(self):
        self.client.close()
        
    def run(self, command, outf=None, errf=None, exception_on_error = True, expectnooutput=False):
        channel = self.client.get_transport().open_session()
        
        log.debug("%s - Running %s" % (self.hostname,command))
        
        if outf == None:
            outf = self.default_outf
    
        if errf == None:
            errf = self.default_errf
            
        try:
            channel.exec_command(command)
            if expectnooutput:
                log.debug("Ignoring output from command (not expecting any)")
            else:
                all_out_nbytes = 0
                all_err_nbytes = 0   
                rem_out = ""
                rem_err = ""
                while True:
                    rl, wl, xl = select.select([channel],[],[], 0.1)
                    if len(rl) > 0:
                        out_nbytes, rem_out = self.__recv(outf, channel.recv_ready, channel.recv, "SSH_OUT", rem_out)
                        err_nbytes, rem_err = self.__recv(errf, channel.recv_stderr_ready, channel.recv_stderr, "SSH_ERR", rem_err)

                        if out_nbytes + err_nbytes == 0:
                            break

                        all_out_nbytes += out_nbytes
                        all_err_nbytes += err_nbytes

                if all_out_nbytes == 0:
                    log.debug("Command did not write to standard output.")

                if all_err_nbytes == 0:
                    log.debug("Command did not write to standard error.")
            
            log.debug("%s - Waiting for exit status: %s" % (self.hostname,command))
            rc = channel.recv_exit_status()
            log.debug("%s - Ran %s (exit status: %i)" % (self.hostname,command,rc))
            channel.close()
        except Exception, e:
            raise # Replace by something more meaningful
         
        if exception_on_error and rc != 0:
            raise SSHCommandFailureException(self, command)
        else:
            return rc
    

        
        
    def scp(self, fromf, tof):
        # Create directory if it does not exist
        try:
            self.sftp.stat(os.path.dirname(tof))
        except IOError, e:
            pdirs = get_parent_directories(tof)
            for d in pdirs:
                try:
                    self.sftp.stat(d)
                except IOError, e:
                    self.sftp.mkdir(d)        
        try:
            self.sftp.put(fromf, tof)
        except Exception, e:
            traceback.print_exc()
            try:
                self.close()
            except:
                pass
        log.debug("scp %s -> %s:%s" % (fromf, self.hostname, tof))
        
    def scp_dir(self, fromdir, todir):
        for root, dirs, files in walk(fromdir):
            todir_full = todir + "/" + root[len(fromdir):]
            try:
                self.sftp.stat(todir_full)
            except IOError, e:
                self.sftp.mkdir(todir_full)
            for f in files:
                fromfile = root + "/" + f
                tofile = todir_full + "/" + f
                self.sftp.put(fromfile, tofile)
                log.debug("scp %s -> %s:%s" % (fromfile, self.hostname, tofile))
                
    def __recv(self, f, ready_func, recv_func, log_label, rem):
        nbytes = 0
        while ready_func():
            data = recv_func(4096)
            if len(data) > 0:
                nbytes += len(data)
                
                if f is not None: 
                    f.write(data)

                lines = data.split('\n')

                if len(lines) == 1:
                    rem += lines[0]
                else:
                    log.debug(log_label + ": %s" % (rem + lines[0]))
                    for line in lines[1:-1]:
                        log.debug(log_label + ": %s" % line)
                    rem = lines[-1]
                
        if f is not None: f.flush()
        
        return nbytes, rem
    

def get_parent_directories(filepath):
    dir = os.path.dirname(filepath)
    dirs = [dir]
    while dir != "/":
        dir = os.path.dirname(dir)
        dirs.append(dir)
    dirs.reverse()
    return dirs
    


from globus.provision.common.utils import rest_table
from globus.provision import AMI, RELEASE

print ".. _ami:"
print

print "Amazon Machine Images for Globus Provision"
print "=========================================="

print """
These are the AMIs currently available for use with Globus Provision.
All these images were generated using the Ubuntu 11.04 AMIs in
`Alestic <http://alestic.com>`_ 

"""

col_names = ["Region", "Architecture", "AMI"]
rows = []
for region in AMI:
    archs = AMI[region].keys()
    archs.sort()
    for arch in archs:
        rows.append((region, arch, AMI[region][arch]))

print rest_table(col_names, rows)

print """
We recommend you use these AMIs when using the EC2 deployer, as these AMIs
include most of the packages required by Globus Provision (you should be able
to use a stock Ubuntu image, but the deployment time will increase considerably,
since those packages will have to be installed before Globus Provision can configure
the instance). 

Take into account that all the sample files in the documentation use the 
32-bit AMI in the us-east-1 region. 
"""
        


# -------------------------------------------------------------------------- #
# Copyright 2010-2011, University of Chicago                                 #
#                                                                            #
# Licensed under the Apache License, Version 2.0 (the "License"); you may    #
# not use this file except in compliance with the License. You may obtain    #
# a copy of the License at                                                   #
#                                                                            #
# http://www.apache.org/licenses/LICENSE-2.0                                 #
#                                                                            #
# Unless required by applicable law or agreed to in writing, software        #
# distributed under the License is distributed on an "AS IS" BASIS,          #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.   #
# See the License for the specific language governing permissions and        #
# limitations under the License.                                             #
# -------------------------------------------------------------------------- #

"""
Instance Management

This module is the single point of access to information about instances.
See the documentation in InstanceStore and Instance for more details.
 
"""

import os.path
import random
from globus.provision.core.config import GPConfig
from globus.provision.core.topology import Topology
from globus.provision.common.certs import CertificateGenerator
from globus.provision.common.persistence import ObjectValidationException

class InstanceException(Exception):
    """A simple exception class used for instance exceptions"""
    pass

class InstanceStore(object):
    """
    The instance database.
    
    Stores information on all the instances created by the user.
    
    Currently, it uses a simple filesystem-based model. There is a
    designated "instances directory" (default: ~/.globusprovision/instances/).
    When an instance is created, a directory with the instance's id is
    created in the instances directory. For example::
    
         ~/.globusprovision/instances/gpi-12345678/
         
    All files related to an instance, including the topology file are
    stored in that directory. Generated files (such as certificates, etc.)
    are also stored there.
    
    The rest of the code only accesses the instances through this class.
    So, it should be possible to eventually replace this with a more elaborate
    solution (e.g., storing the instance data in a database, etc.)    
    """
    
    def __init__(self, instances_dir):
        self.instances_dir = instances_dir
        
    def create_new_instance(self, topology_json, config_txt):
        created = False
        while not created:
            inst_id = "gpi-" + hex(random.randint(1,2**31-1))[2:].rjust(8,"0")
            inst_dir = "%s/%s" % (self.instances_dir, inst_id)
            if not os.path.exists(inst_dir):
                os.makedirs(inst_dir)
                created = True
                
        configf = open("%s/provision.conf" % inst_dir, "w")
        configf.write(config_txt)
        configf.close()

        # We don't do anything with it. Just use it to raise an exception
        # if there is anything wrong with the configuration file
        GPConfig("%s/provision.conf" % inst_dir)

        topology = Topology.from_json_string(topology_json)
        topology.set_property("id", inst_id)
        topology.set_property("state", Topology.STATE_NEW)
        topology.save("%s/topology.json" % inst_dir)
                                        
        inst = Instance(inst_id, inst_dir)
        
        return inst
    
    def get_instance(self, inst_id):
        inst_dir = "%s/%s" % (self.instances_dir, inst_id)

        if not os.path.exists(inst_dir):
            raise InstanceException("Instance %s does not exist" % inst_id)
        return Instance(inst_id, inst_dir)

    def get_instances(self, inst_ids = None):
        valid_instances = []
        invalid_instances = []
        
        for inst_id in self.__get_instance_ids():
            if inst_ids == None or (inst_ids != None and inst_id in inst_ids):
                try:
                    inst = Instance(inst_id, "%s/%s" % (self.instances_dir, inst_id))
                    valid_instances.append(inst)
                except Exception, e:
                    invalid_instances.append((inst_id,str(e)))
        
        return (valid_instances, invalid_instances)
    
    def __get_instance_ids(self):
        inst_ids = [i for i in os.listdir(self.instances_dir)]
        return inst_ids

class Instance(object):
    """
    A Globus Provision Instance
    
    This class represents a single instance. Right now, an instance is 
    the combination of a configuration file and a topology (both of which
    are provided when the instance is created).
    
    The configuration file contains all the information about
    the instance that will (arguably) not change during its lifetime.
    The topology contains the specification of the hosts, users, etc.
    that are going to be deployed, and that could change during the
    instance's lifetime.
    
    For example, the configuration file specifies what keypair to use
    when accessing EC2. Although this could conceivably change, it is
    not as likely as a change in the topology (e.g., adding a new host,
    changing the run list of a host, etc.)
    
    """
    
    # Relative to instance directory
    CERTS_DIR = "/certs"

    def __init__(self, inst_id, instance_dir):
        self.instance_dir = instance_dir
        self.id = inst_id
        self.config = GPConfig("%s/provision.conf" % instance_dir)
        self.topology = self.__load_topology()
        
    def __load_topology(self):
        topology_file = "%s/topology.json" % self.instance_dir
        f = open (topology_file, "r")
        json_string = f.read()
        topology = Topology.from_json_string(json_string)
        topology._json_file = topology_file
        f.close()   
        return topology     

    def update_topology(self, topology_json):
        try:
            topology_file = "%s/topology.json" % self.instance_dir        
            new_topology = Topology.from_json_string(topology_json)
            new_topology._json_file = topology_file
        except ObjectValidationException, ove:
            message = "Error in topology file: %s" % ove 
            return (False, message, None)

        try:
            topology_changes = self.topology.validate_update(new_topology)
        except ObjectValidationException, ove:
            message = "Could not update topology: %s" % ove 
            return (False, message, None)

        self.topology = new_topology
        self.topology.save()
        
        return (True, "Success", topology_changes)

    def gen_certificates(self, force_hosts = False, force_users = False, force_ca = False):
        certs_dir = self.instance_dir + self.CERTS_DIR
        if not os.path.exists(certs_dir):
            os.makedirs(certs_dir)  

        dn = self.config.get("ca-dn")
        if dn == None:
            dn = "O=Grid, OU=Globus Provision (generated)"

        certg = CertificateGenerator(dn)

        cert_files = []
        ca_cert_file = self.config.get("ca-cert")
        ca_cert_key = self.config.get("ca-key")
        
        if ca_cert_file != None:
            ca_cert, ca_key = certg.load_certificate(ca_cert_file, ca_cert_key)
        else:
            ca_cert, ca_key = certg.gen_selfsigned_ca_cert("Globus Provision CA")
        
        certg.set_ca(ca_cert, ca_key)

        h = "%x" % ca_cert.subject_name_hash()

        hash_file = open(certs_dir + "/ca_cert.hash", "w")
        hash_file.write(h)
        hash_file.close()   

        ca_cert_file = "%s/ca_cert.pem" % certs_dir
        ca_key_file = certs_dir + "/ca_key.pem"
        cert_files.append(ca_cert_file)
        cert_files.append(ca_key_file)
        certg.save_certificate(cert = ca_cert,
                              key = ca_key,
                              cert_file = ca_cert_file,
                              key_file = ca_key_file, 
                              force = force_ca)

        users = [u for u in self.topology.get_users() if u.certificate=="generated"]
        for user in users:        
            cert, key = certg.gen_user_cert(cn = user.id) 
            
            cert_file = "%s/%s_cert.pem" % (certs_dir, user.id)
            key_file = "%s/%s_key.pem" % (certs_dir, user.id)
            cert_files.append(cert_file)
            cert_files.append(key_file)    
            certg.save_certificate(cert = cert,
                                    key = key,
                                    cert_file = cert_file,
                                    key_file = key_file, 
                                    force = force_users)
        
        nodes = self.topology.get_nodes()
        for n in nodes:
            cert, key = certg.gen_host_cert(hostname = n.hostname) 
            
            filename = n.id
            
            cert_file = "%s/%s_cert.pem" % (certs_dir, filename)
            key_file = "%s/%s_key.pem" % (certs_dir, filename)
            cert_files.append(cert_file)
            cert_files.append(key_file)          
            certg.save_certificate(cert = cert,
                                   key = key,
                                   cert_file = cert_file,
                                   key_file = key_file, 
                                   force = force_hosts)        

        return cert_files  


                 
        


        



# -------------------------------------------------------------------------- #
# Copyright 2010-2011, University of Chicago                                 #
#                                                                            #
# Licensed under the Apache License, Version 2.0 (the "License"); you may    #
# not use this file except in compliance with the License. You may obtain    #
# a copy of the License at                                                   #
#                                                                            #
# http://www.apache.org/licenses/LICENSE-2.0                                 #
#                                                                            #
# Unless required by applicable law or agreed to in writing, software        #
# distributed under the License is distributed on an "AS IS" BASIS,          #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.   #
# See the License for the specific language governing permissions and        #
# limitations under the License.                                             #
# -------------------------------------------------------------------------- #

"""
Commands related to the EC2 deployer, but which do not require access to the API
"""

import sys
import os.path

from globus.provision.cli import Command
from globus.provision.deploy.ec2.images import EC2AMICreator, EC2AMIUpdater
from globus.provision.common import defaults
from globus.provision.core.config import GPConfig
from globus.provision.common.utils import parse_extra_files_files

        
def gp_ec2_create_ami_func():
    return gp_ec2_create_ami(sys.argv).run()        

class gp_ec2_create_ami(Command):
    """
    Creates a Globus Provision AMI with Chef files pre-deployed, and
    some software pre-installed.            
    """
    
    name = "gp-ec2-create-ami"
    
    def __init__(self, argv):
        Command.__init__(self, argv)
        
        self.optparser.add_option("-s", "--chef-directory", 
                                  action="store", type="string", dest="chef_dir", 
                                  help = "Location of Chef files.")        
        
        self.optparser.add_option("-c", "--conf", 
                                  action="store", type="string", dest="conf", 
                                  default = defaults.CONFIG_FILE,
                                  help = "Configuration file. Must include an [ec2] section.") 

        self.optparser.add_option("-a", "--ami", 
                                  action="store", type="string", dest="ami", 
                                  help = "AMI to base the new AMI on.")

        self.optparser.add_option("-n", "--name", 
                                  action="store", type="string", dest="aminame", 
                                  help = "Name of AMI to create")

        self.optparser.add_option("-t", "--instance-type", 
                                  action="store", type="string", dest="instance_type", 
                                  help = "EC2 instance type to use")
                
    def run(self):    
        self.parse_options()

        config = GPConfig(os.path.expanduser(self.opt.conf))
        chef_dir = os.path.expanduser(self.opt.chef_dir)

        c = EC2AMICreator(chef_dir, self.opt.ami, self.opt.aminame, self.opt.instance_type, config)
        c.run()
        
        
def gp_ec2_update_ami_func():
    return gp_ec2_update_ami(sys.argv).run()     
        
class gp_ec2_update_ami(Command):
    """
    Takes an existing AMI, adds files to it, and creates a new AMI.
    """    
    
    name = "gp-ec2-update-ami"
    
    def __init__(self, argv):
        Command.__init__(self, argv)
        
        self.optparser.add_option("-a", "--ami", 
                                  action="store", type="string", dest="ami", 
                                  help = "AMI to update.")

        self.optparser.add_option("-n", "--name", 
                                  action="store", type="string", dest="aminame", 
                                  help = "Name of new AMI.")

        self.optparser.add_option("-c", "--conf", 
                                  action="store", type="string", dest="conf", 
                                  default = defaults.CONFIG_FILE,
                                  help = "Configuration file. Must include an [ec2] section.") 
        
        self.optparser.add_option("-f", "--files", 
                                  action="store", type="string", dest="files", 
                                  help = "Files to add to AMI")
                
    def run(self):    
        self.parse_options()
        
        files = parse_extra_files_files(self.opt.files)
        config = GPConfig(os.path.expanduser(self.opt.conf))
        
        c = EC2AMIUpdater(self.opt.ami, self.opt.aminame, files, config)
        c.run()        




# -------------------------------------------------------------------------- #
# Copyright 2010-2011, University of Chicago                                 #
#                                                                            #
# Licensed under the Apache License, Version 2.0 (the "License"); you may    #
# not use this file except in compliance with the License. You may obtain    #
# a copy of the License at                                                   #
#                                                                            #
# http://www.apache.org/licenses/LICENSE-2.0                                 #
#                                                                            #
# Unless required by applicable law or agreed to in writing, software        #
# distributed under the License is distributed on an "AS IS" BASIS,          #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.   #
# See the License for the specific language governing permissions and        #
# limitations under the License.                                             #
# -------------------------------------------------------------------------- #

"""
Contains the parsers for the two configuration files used in Globus Provision:

* The instance configuration file (GPConfig): This is the configuration file
  that specifies options related to an instance's deploymenr.
  
* The simple topology file: This is a simple format for specifying topologies
  (which internally translated to the topology JSON format). It has the
  format of a configuration file although, strictly speaking, it is *not*
  a configuration file.    

"""

from globus.provision.core.topology import Domain, User, Node, Topology,\
    DeployData, EC2DeployData, GridMapEntry, GOEndpoint
from globus.provision.common.config import Config, Section, Option, OPTTYPE_INT, OPTTYPE_FLOAT, OPTTYPE_STRING, OPTTYPE_BOOLEAN, OPTTYPE_FILE
import os.path
import getpass

class GPConfig(Config):
    """
    The instance configuration file.
    """

    sections = []    
    
    # ============================= #
    #                               #
    #        GENERAL OPTIONS        #
    #                               #
    # ============================= #    
    
    general = Section("general", required=True,
                      doc = "This section is used for general options affecting Globus Provision as a whole.")
    general.options = \
    [
     Option(name        = "ca-cert",
            getter      = "ca-cert",
            type        = OPTTYPE_FILE,
            required    = False,
            doc         = """
            Location of CA certificate (PEM-encoded) used to generate user
            and host certificates. If blank, Globus Provision will generate a self-signed
            certificate from scratch.        
            """),    
     Option(name        = "ca-key",
            getter      = "ca-key",
            type        = OPTTYPE_FILE,
            required    = False,
            doc         = """
            Location of the private key (PEM-encoded) for the certificate
            specified in ``ca-cert``.
            """),
     Option(name        = "ca-dn",
            getter      = "ca-dn",
            type        = OPTTYPE_STRING,
            required    = False,
            doc         = """
            Distinguished Name of the certificates that will be signed with 
            the CA certificate specified in ``ca-cert``. 
            
            For example, if you set this value to ``O=Foo, OU=Bar``, the certificates
            will have subjects like ``/O=Foo/OU=Bar/CN=borja``, ``/O=Foo/OU=Bar/CN=host/foo.example.org``, etc.
            """),            
     Option(name        = "scratch-dir",
            getter      = "scratch-dir",
            type        = OPTTYPE_STRING,
            required    = False,
            default     = "/var/tmp",
            doc         = """
            Scratch directory that Chef will use (on the provisioned machines)
            while configuring them.
            """),
     Option(name        = "deploy",
            getter      = "deploy",
            type        = OPTTYPE_STRING,
            required    = True,
            valid       = ["ec2", "dummy"],
            doc         = """
            Globus Provision can support various "deployers" that are used to
            deploy the hosts in a topology. Two deployers are currently supported:
            
            * ``ec2``: Hosts are deployed as Amazon EC2 instances.
            * ``dummy``: Hosts are not actually deployed and are assigned dummy
              hostnames and IP addresses.
              
            See the Globus Provision documentation for more details on the
            available deployers.
            """)     
    ]

    sections.append(general)

    # ====================== #
    #                        #
    #      EC2 OPTIONS       #
    #                        #
    # ====================== #

    ec2 = Section("ec2", required=False,
                         required_if = [(("general","deploy"),"ec2")],
                         doc = """
                         When the EC2 deployer is selected, Globus Provision will need certain information about
                         your EC2 account to be able to request EC2 instances on which to deploy your topology. This account
                         information is specified in this section of the configuration file. If you are unclear on what values
                         you need to specify here, see :ref:`chap_ec2` for more detailed instructions (including how to set up
                         an Amazon EC2 account)""")
    ec2.options = \
    [                
     Option(name        = "keypair",
            getter      = "ec2-keypair",
            type        = OPTTYPE_STRING,
            required    = True,
            doc         = """
            The *name* of the Amazon EC2 keypair you will use to log into the VMs.
            See :ref:`chap_ec2` for instructions on how to obtain this keypair.
            """),
     Option(name        = "keyfile",
            getter      = "ec2-keyfile",
            type        = OPTTYPE_FILE,
            required    = True,
            doc         = """
            The actual location of the keypair on your local filesystem.
            See :ref:`chap_ec2` for instructions on how to obtain this keypair.
            """),
     Option(name        = "username",
            getter      = "ec2-username",
            type        = OPTTYPE_STRING,
            required    = True,
            doc         = """
            The username that Globus Provision will use to connect to the EC2 instances,
            using the keypair specified in ``keypair``. If you are using one of the
            Globus Provision AMIs, you need to set this value to ``ubuntu``.
            """),     
     Option(name        = "server-hostname",
            getter      = "ec2-server-hostname",
            type        = OPTTYPE_STRING,
            required    = False,
            doc         = """
            The hostname of the EC2 server. If you are using Amazon AWS, leave this option
            unspecified. If you are using an EC2-compatible system, such as OpenNebula, Nimbus,
            Eucalyptus, etc. set this to the server running that system's EC2 interface.
            """),
     Option(name        = "server-port",
            getter      = "ec2-server-port",
            type        = OPTTYPE_INT,
            required    = False,
            doc         = """
            The TCP port of the EC2 server. If you are using Amazon AWS, leave this option
            unspecified. If you are using an EC2-compatible system, such as OpenNebula, Nimbus,
            Eucalyptus, etc. set this to the port on which that system's EC2 interface is listening on.
            """),
     Option(name        = "server-path",
            getter      = "ec2-server-path",
            type        = OPTTYPE_STRING,
            required    = False,
            doc         = """
            The path portion of the EC2 server. If you are using Amazon AWS, leave this option
            unspecified. If you are using an EC2-compatible system, such as OpenNebula, OpenStack,
            Eucalyptus, etc. set this to the path (in the host specified in ``server-hostname``)
            that the system's EC2 interface is available on.
            """)                                    
    ]    
    sections.append(ec2)

    # ================================ #
    #                                  #
    #      GLOBUS ONLINE OPTIONS       #
    #                                  #
    # ================================ #

    go = Section("globusonline", required=False,
                  doc = """
                  When a topology includes Globus Online transfer endpoints, Globus Provision will
                  use GO's API to set up those endpoints. To do so, it will need some information
                  about your GO account. If you are unclear on what values you need to specify here, 
                  see :ref:`chap_go` for more detailed instructions.
                  """)
    go.options = \
    [       
   
     Option(name        = "ssh-key",
            getter      = "go-ssh-key",
            type        = OPTTYPE_FILE,
            default     = "~/.ssh/id_rsa",            
            required    = False,
            doc         = """
            SSH key to use when connecting to the Globus Online CLI. The public key
            for this SSH key must have been previously added to your Globus Online
            profile.
            """),          
     Option(name        = "cert-file",
            getter      = "go-cert-file",
            type        = OPTTYPE_FILE,
            required    = False,
            doc         = """
            When this option is specified, Globus Provision will access your GO
            account using Globus Online's Transfer API (instead of sending commands
            to Globus Online's CLI via SSH). To do so, Globus Provision needs the
            location of a user certificate (PEM-encoded) that is authorized to access 
            the accounts specified in your topology's endpoints.
            
            See :ref:`chap_go` for more details on the differences between using the
            Transfer API, instead of the CLI via SSH.
            """),         
     Option(name        = "key-file",
            getter      = "go-key-file",
            type        = OPTTYPE_FILE,
            required    = False,
            doc         = """
            Location of the private key (PEM-encoded) for the certificate
            specified in ``cert-file``.
            """),        
     Option(name        = "server-ca-file",
            getter      = "go-server-ca-file",
            type        = OPTTYPE_STRING,
            required    = False,
            doc         = """
            To verify the server certificate of the Globus Online Transfer API server,
            Globus Provision needs the certificate of the CA that signed that certificate.
            This file is already bundled with Globus Provision. The only reason for using
            this option to specify a different CA certificate is in the unlikely case that
            the API server decides to switch to a different CA (and the file bundled
            with Globus Provision has not been updated to that CA yet).
            """)
    ]         
    sections.append(go)
      
    def __init__(self, config_file):
        Config.__init__(self, config_file, self.sections)


class SimpleTopologyConfig(Config):
    """
    The simple topology file
    """    
    
    sections = []    
    
    # ============================= #
    #                               #
    #        GENERAL OPTIONS        #
    #                               #
    # ============================= #    
    
    general = Section("general", required=True,
                      doc = "This section is used for general options affecting all the topology.")
    general.options = \
    [    
     Option(name        = "domains",
            getter      = "domains",
            type        = OPTTYPE_STRING,
            required    = True,
            doc         = """
            The names of the domains you are defining in this topology. They must each be separated by
            a single space.      
            """),    
     Option(name        = "deploy",
            getter      = "deploy",
            type        = OPTTYPE_STRING,
            required    = True,
            valid       = ["ec2", "dummy"],
            doc         = """
            See the :ref:`deploy option <GPConfig_deploy>` in :ref:`chap_config_ref` 
            """),
     Option(name        = "ssh-pubkey",
            getter      = "ssh-pubkey",
            type        = OPTTYPE_FILE,
            required    = False,
            default     = "~/.ssh/id_rsa.pub",
            doc         = """
            When creating users, an SSH public key must be added to their ``authorized_keys`` file
            to allow the creator of the topology to log in as those users. When using a topology file,
            each SSH key is specified separately for each user; in a simple topology, you can specify
            a single SSH public key for all the users (by default, the SSH key of the topology's creator 
            will be used)
            
            Take into account that you *can* specify per-user SSH keys in a simple topology by using the
            :ref:`users-file option<SimpleTopologyConfig_users-file>`.
            """)                                   
    ]     
    
    sections.append(general)
    
    domain = Section("domain", required=False, multiple=("general", "domains"),
                     doc = """
                     For each domain specified in the ``domains`` option, you will need to specify a section
                     titled ``[domain-DDD]`` where ``DD`` is the name of the domain. For example, if you specify the following::
                     
                         [general]
                         domains: foo bar
                         
                     You will need to specify the following sections::
                     
                         [domain-foo]
                         ...
                         
                         [domain-bar]
                         ...
                         
                     Each section provides a few high-level options about each domain.
                     This provides a simple, but constrained, way of specifying what services and users
                     should be created in each domain. For more complex topologies, you may have
                     to write a regular :ref:`topology file <chap_topology>`.                     
                     """)
    domain.options = \
    [    
     Option(name        = "users",
            getter      = "users",
            type        = OPTTYPE_STRING,
            required    = False,
            default     = "0",
            doc         = """
            This option can be either a number, or a list of usernames separated by spaces.
            
            If a number is specified, the users will be named ``D-userN``, where ``D`` is the
            domain name and ``N`` is a number between 1 and the number specified in this option.
            
            If a list of usernames is specified, users with those login names will be created.
            
            These users will be created with corresponding user certificates. To create users without user certificates
            use option ``users-no-cert``.        
            """),                             
     Option(name        = "users-no-cert",
            getter      = "users-no-cert",
            type        = OPTTYPE_STRING,
            default     = "0",
            required    = False,
            doc         = """
            Same as ``users`` but creating users without certificates.
            
            Note that if you specify a number for *both* the ``users`` and ``users-no-cert`` option 
            (with values N and M, respectively), the first N users will have certificates, and the 
            remaining M will not.        
            """),                   
     Option(name        = "users-file",
            getter      = "users-file",
            type        = OPTTYPE_FILE,
            required    = False,
            doc         = """
            The path to a file with a specification of the users to create in this domain. This file will have one line
            per user, each with three fields (separated by whitespace):
            
            #. A single character, ``C`` or ``N``. If ``C`` is specified, the user will have a user certificate created
               for it. Otherwise, it will nor.
            #. The user's UNIX login name.
            #. (Optional) An SSH public key to add to the user's ``authorized_keys`` file. If not specified, the public
               key specified in :ref:`option ssh-pubkey<SimpleTopologyConfig_ssh-pubkey>` will be used.
              
            For example::
            
                C borja     ssh-rsa FOOFOOFOO...BARBARBAR borja@example.org
                C childers  ssh-rsa FOOFOOFOO...BARBARBAR childers@example.org
                N foster
                N madduri
            
            """),   
     
     Option(name        = "barebones-nodes",
            getter      = "barebones-nodes",
            type        = OPTTYPE_INT,
            default     = 0,
            required    = False,
            doc         = """
            A "barebones node" is a node on which no software will be installed. If ``nfs-nis`` is ``True``,
            these nodes *will* be configured as NFS/NIS clients. These nodes can be useful for testing.   
            """),            
     
     Option(name        = "nfs-nis",
            getter      = "nfs-nis",
            type        = OPTTYPE_BOOLEAN,
            required    = False,
            default     = False,
            doc         = """
            Specifies whether an NFS/NIS server should be setup in this domain. When ``True``, there will be a global
            filesystem and global user account space in the domain. Most notably, the users' home directories will be on an
            NFS directory, which means they will be able to access the same home directory from any host in the domain
            (as opposed to having separate home directories in each host).
            
            When ``False``, user accounts and home directories will be created on every individual host. This option can
            be useful if you are creating a single-host domain.       
            """),             
     Option(name        = "login",
            getter      = "login",
            type        = OPTTYPE_BOOLEAN,
            required    = False,
            default     = False,
            doc         = """
            Specifies whether a separate "login node" should be created in the topology. This option can be useful if you
            want a distinct node that users can log into but that does not host one of the topology's servers (like the NFS
            server, a GridFTP server, etc.)        
            """),
     Option(name        = "myproxy",
            getter      = "myproxy",
            type        = OPTTYPE_BOOLEAN,
            required    = False,
            default     = False,
            doc         = """
            Specifies whether to set up a MyProxy server on this domain.        
            """),                           
     Option(name        = "gram",
            getter      = "gram",
            type        = OPTTYPE_BOOLEAN,
            required    = False,
            default     = False,
            doc         = """
            Specifies whether to set up a GRAM5 server on this domain.             
            """),                   
     Option(name        = "gridftp",
            getter      = "gridftp",
            type        = OPTTYPE_BOOLEAN,
            required    = False,
            default     = False,
            doc         = """
            Specifies whether to set up a GridFTP server on this domain.             
            """),                   
     Option(name        = "lrm",
            getter      = "lrm",
            type        = OPTTYPE_STRING,
            valid       = ["none", "condor"],
            default     = "none",
            required    = False,
            doc         = """
            Specifies whether to set up an LRM (Local Resource Manager) on this domain. Currently, only          
            `Condor <http://www.cs.wisc.edu/condor/>`_ is supported.   
            """),           
     Option(name        = "cluster-nodes",
            getter      = "cluster-nodes",
            type        = OPTTYPE_INT,
            required    = False,
            doc         = """
            The number of worker nodes to create for the LRM.        
            """),         
     Option(name        = "galaxy",
            getter      = "galaxy",
            type        = OPTTYPE_BOOLEAN,
            required    = False,
            default     = False,
            doc         = """
            Specifies whether to set up a Galaxy server on this domain.        
            """),   
              
     Option(name        = "go-endpoint",
            getter      = "go-endpoint",
            type        = OPTTYPE_STRING,
            required    = False,
            doc         = """
            If this domain has a GridFTP server, it can be configured as a GO endpoint.
            The format for this option is <username>#<name> (e.g., johnsmith#test-ep).
            Take into account that you must be authorized to use the GO account for <username>,
            and that you must specify the appropriate credentials in the 
            :ref:`[globusonline] section<GPConfig_section_globusonline>` of the configuration file.
            
            See :ref:`chap_go` for more details.    
            """),
               
     Option(name        = "go-auth",
            getter      = "go-auth",
            type        = OPTTYPE_STRING,
            required    = False,
            valid       = ["myproxy", "go"],            
            doc         = """
            The authentication method that Globus Online will use when contacting the endpoint on
            behalf of a user. The valid options are:
            
            * ``myproxy``: Contact the MyProxy server specified in the topology. Note that 
              the :ref:`myproxy option<SimpleTopologyConfig_myproxy>` must be set to ``true`` 
              for this to work
            * ``go``: Use Globus Online authentication.
              
            See :ref:`chap_go`, and specifically :ref:`Globus Online Authentication Methods <sec_go_auth>`,
            for more details on the implications of each authentication method.            
            """),
     
     Option(name        = "go-gc",
            getter      = "go-gc",
            type        = OPTTYPE_BOOLEAN,
            required    = False,
            default     = True,
            doc         = """
            If true, the endpoint will use a Globus Connect certificate.
            If not, it will use the host certificate generated by Globus Provision.
            Take into account that, for the GridFTP server to be trusted by
            Globus Online, it must use a certificate trusted by Globus Online.
            Unless you used a CA trusted by Globus Online to generate the certificates
            for the topology, you must use a Globus Connect certificate.
            """)                            
    ]     
    sections.append(domain)
    
    ec2 = Section("ec2", required=False,
                         required_if = [(("general","deploy"),"ec2")],
                         doc = """
                         When the EC2 deployer is selected, this section will allow you to
                         specify EC2 deployment options that are specific to this topology.""")    
    ec2.options = \
    [                
     Option(name        = "ami",
            getter      = "ec2-ami",
            type        = OPTTYPE_STRING,
            required    = True,
            doc         = """
            This is the AMI (`Amazon Machine Image <http://en.wikipedia.org/wiki/Amazon_Machine_Image>`_) 
            that Globus Provision will use to create each host in the domani. Any recent Ubuntu or Debian
            AMI should work. Nonetheless, take into account that we provide an AMI that has most of the
            necessary software pre-installed in it, considerably speeding up the setup of the machines. 
            The latest Globus Provision AMI is always listed in the Globus Provision website.
            """),
     Option(name        = "instance-type",
            getter      = "ec2-instance-type",
            type        = OPTTYPE_STRING,
            required    = True,
            default     = "t1.micro",
            doc         = """
            This is the `EC2 instance type <http://en.wikipedia.org/wiki/Amazon_Machine_Image>`_ that will
            be used to launch the machines in this domain. The default is to use micro-instances (t1.micro),
            which tend to be enough if you are just tinkering around.
            """),         
     Option(name        = "availability-zone",
            getter      = "ec2-availability-zone",
            type        = OPTTYPE_STRING,
            required    = False,
            default     = None,
            doc         = """
            The `availability zone <http://docs.amazonwebservices.com/AWSEC2/latest/UserGuide/concepts-regions-availability-zones.html>`_ 
            you want the VMs to be deployed in. 
            Unless you have a good reason for choosing a specific availability zone,
            you should let Globus Provision choose a default zone for you.
            """)          
    ]    
    sections.append(ec2)    
  
 
    def __init__(self, configfile):
        Config.__init__(self, configfile, self.sections)

    def to_topology(self):
        ssh_pubkeyf = os.path.expanduser(self.get("ssh-pubkey"))
        ssh_pubkeyf = open(ssh_pubkeyf)
        ssh_pubkey = ssh_pubkeyf.read().strip()
        ssh_pubkeyf.close()        
        
        topology = Topology()
        
        if self.get("deploy") == "dummy":
            # No default deploy data
            pass
        elif self.get("deploy") == "ec2":
            deploy_data = DeployData()
            ec2_deploy_data = EC2DeployData()
            
            ec2_deploy_data.set_property("ami", self.get("ec2-ami"))
            ec2_deploy_data.set_property("instance_type", self.get("ec2-instance-type"))
            
            deploy_data.set_property("ec2", ec2_deploy_data)
            topology.set_property("default_deploy_data", deploy_data)
        
        domains = self.get("domains").split()
        for domain_name in domains:
            domain = Domain()
            domain.set_property("id", domain_name)
            topology.add_to_array("domains", domain)

            has_go_ep = self.get((domain_name,"go-endpoint")) != None

            user = User()
            user.set_property("id", getpass.getuser())
            user.set_property("password_hash", "!")
            user.set_property("certificate", "generated")
            user.set_property("admin", True)
            user.set_property("ssh_pkey", ssh_pubkey)
            domain.add_user(user)            

            usersfile = self.get((domain_name, "users-file"))
            
            if usersfile != None:
                usersfile = open(usersfile, "r")
                
                for line in usersfile:
                    fields = line.split()
                    type = fields[0]
                    username = fields[1]
                    if len(fields) >= 3:
                        user_ssh_pubkey = " ".join(fields[2:])
                    else:
                        user_ssh_pubkey = ssh_pubkey
                    
                    user = User()
                    user.set_property("id", username)
                    user.set_property("password_hash", "!")
                    user.set_property("ssh_pkey", user_ssh_pubkey)
                    if type == "C":
                        user.set_property("certificate", "generated")
                    else:
                        user.set_property("certificate", "none")
                        
                    domain.add_user(user)
                    
                usersfile.close()
            else:
                users = self.get((domain_name, "users"))
                users_nocert = self.get((domain_name, "users-no-cert"))
                
                if users.isdigit():
                    num_users = int(users)
                    usernames = [("%s-user%i" % (domain_name, i), True) for i in range(1,num_users + 1)]
                else:
                    num_users = 0
                    usernames = [(u, True) for u in users.split() if u != getpass.getuser()]
                    
                if users_nocert.isdigit():
                    usernames += [("%s-user%i" % (domain_name, i), False) for i in range(num_users + 1,num_users + int(users_nocert) + 1)]
                else:
                    usernames += [(u, False) for u in users_nocert.split() if u != getpass.getuser()]                

                for username, cert in usernames:
                    user = User()
                    user.set_property("id", username)
                    user.set_property("password_hash", "!")
                    user.set_property("ssh_pkey", ssh_pubkey)
                    if cert:
                        user.set_property("certificate", "generated")
                    else:
                        user.set_property("certificate", "none")
                    domain.add_user(user)
                
            for user in domain.users.values():
                gme = GridMapEntry()
                gme.set_property("dn", "/O=Grid/OU=Globus Provision (generated)/CN=%s" % user.id)
                gme.set_property("login", user.id)
                domain.add_to_array("gridmap", gme)  
                if self.get((domain_name,"go-auth")) == "go":
                    gme = GridMapEntry()
                    gme.set_property("dn", "/C=US/O=Globus Consortium/OU=Globus Connect User/CN=%s" % user.id)
                    gme.set_property("login", user.id)
                    domain.add_to_array("gridmap", gme)  
                                  
            
            if self.get((domain_name,"nfs-nis")):  
                server_node = Node()
                server_name = "%s-server" % domain_name
                server_node.set_property("id", server_name)
                server_node.add_to_array("run_list", "role[domain-nfsnis]")
                if not self.get((domain_name,"login")):
                    # If there is no login node, the NFS/NIS server will
                    # effectively act as one. 
                    server_node.add_to_array("run_list", "role[globus]")
                if self.get((domain_name,"galaxy")):
                    # If there is a Galaxy server in the domain, the "common"
                    # recipe has to be installed on the NFS/NIS server
                    server_node.add_to_array("run_list", "recipe[galaxy::galaxy-globus-common]")
                    
                domain.add_node(server_node)

            for i in range(self.get((domain_name,"barebones-nodes"))):
                bb_name = "%s-blank-%i" % (domain_name, i+1)

                bb_node = Node()
                bb_node.set_property("id", bb_name)
                if self.get((domain_name,"nfs-nis")):  
                    bb_node.set_property("depends", "node:%s" % server_name)
                    bb_node.add_to_array("run_list", "role[domain-nfsnis-client]")
                else:
                    bb_node.add_to_array("run_list", "recipe[provision::gp_node]")  
                    bb_node.add_to_array("run_list", "recipe[provision::domain_users]")              
                               
                domain.add_node(bb_node)

            if self.get((domain_name,"login")):            
                login_node = Node()
                login_node.set_property("id", "%s-login" % domain_name)
                if self.get((domain_name,"nfs-nis")):  
                    login_node.set_property("depends", "node:%s" % server_name)
                    login_node.add_to_array("run_list", "role[domain-nfsnis-client]")
                else:
                    login_node.add_to_array("run_list", "recipe[provision::gp_node]")
                    login_node.add_to_array("run_list", "recipe[provision::domain_users]")
                login_node.add_to_array("run_list", "role[globus]")
                domain.add_node(login_node)                

            if self.get((domain_name,"myproxy")):
                myproxy_node = Node()
                myproxy_node.set_property("id", "%s-myproxy" % domain_name)
                if self.get((domain_name,"nfs-nis")):  
                    myproxy_node.set_property("depends", "node:%s" % server_name)
                    myproxy_node.add_to_array("run_list", "role[domain-nfsnis-client]")
                else:
                    myproxy_node.add_to_array("run_list", "recipe[provision::gp_node]")
                    myproxy_node.add_to_array("run_list", "recipe[provision::domain_users]")
                myproxy_node.add_to_array("run_list", "role[domain-myproxy]")
                domain.add_node(myproxy_node)

            if self.get((domain_name,"gridftp")):
                gridftp_node = Node()
                gridftp_node.set_property("id", "%s-gridftp" % domain_name)
                if self.get((domain_name,"nfs-nis")):  
                    gridftp_node.set_property("depends", "node:%s" % server_name)
                    gridftp_node.add_to_array("run_list", "role[domain-nfsnis-client]")
                else:
                    gridftp_node.add_to_array("run_list", "recipe[provision::gp_node]")
                    gridftp_node.add_to_array("run_list", "recipe[provision::domain_users]")     
                    
                if has_go_ep:
                    if self.get((domain_name,"go-gc")):
                        gridftp_node.add_to_array("run_list", "role[domain-gridftp-gc]")
                    else:
                        gridftp_node.add_to_array("run_list", "recipe[globus::go_cert]")
                        gridftp_node.add_to_array("run_list", "role[domain-gridftp-default]")
                else:                
                    gridftp_node.add_to_array("run_list", "role[domain-gridftp-default]")
                domain.add_node(gridftp_node)                
            
            if self.get((domain_name,"galaxy")):
                galaxy_node = Node()
                galaxy_node.set_property("id", "%s-galaxy" % domain_name)

                if self.get((domain_name,"nfs-nis")):  
                    galaxy_node.set_property("depends", "node:%s" % server_name)
                    galaxy_node.add_to_array("run_list", "role[domain-nfsnis-client]")
                else:
                    galaxy_node.add_to_array("run_list", "recipe[provision::gp_node]")
                    galaxy_node.add_to_array("run_list", "recipe[provision::domain_users]")     
                    galaxy_node.add_to_array("run_list", "recipe[galaxy::galaxy-globus-common]")     

                if self.get((domain_name,"go-endpoint")) != None:
                    galaxy_node.add_to_array("run_list", "recipe[globus::go_cert]")
                galaxy_node.add_to_array("run_list", "recipe[galaxy::galaxy-globus]")
                domain.add_node(galaxy_node)                
            
            
            lrm = self.get((domain_name,"lrm"))
            if lrm != "none":
                gram = self.get((domain_name,"gram"))
                if lrm == "condor":
                    if gram:
                        node_name = "%s-gram-condor" % domain_name
                        role = "role[domain-gram-condor]"
                    else:
                        node_name = "%s-condor" % domain_name
                        role = "role[domain-condor]"
                    workernode_role = "role[domain-clusternode-condor]"

                lrm_node = Node()
                lrm_node.set_property("id", node_name)
                if self.get((domain_name,"nfs-nis")):  
                    lrm_node.set_property("depends", "node:%s" % server_name)
                    lrm_node.add_to_array("run_list", "role[domain-nfsnis-client]")
                else:
                    lrm_node.add_to_array("run_list", "recipe[provision::gp_node]")
                    lrm_node.add_to_array("run_list", "recipe[provision::domain_users]")                    
                lrm_node.add_to_array("run_list", role)
                domain.add_node(lrm_node)

                clusternode_host = 1
                for i in range(self.get((domain_name,"cluster-nodes"))):
                    wn_name = "%s-condor-wn%i" % (domain_name, i+1)

                    wn_node = Node()
                    wn_node.set_property("id", wn_name)
                    wn_node.set_property("depends", "node:%s" % node_name)
                    if self.get((domain_name,"nfs-nis")):
                        wn_node.add_to_array("run_list", "role[domain-nfsnis-client]")          
                    else:
                        wn_node.add_to_array("run_list", "recipe[provision::gp_node]")
                        wn_node.add_to_array("run_list", "recipe[provision::domain_users]")                                  
                    wn_node.add_to_array("run_list", workernode_role)
                    domain.add_node(wn_node)

                    clusternode_host += 1
            
            if has_go_ep:
                goep = GOEndpoint()
                gouser, goname = self.get((domain_name,"go-endpoint")).split("#")
                goep.set_property("user", gouser)
                goep.set_property("name", goname)
                goep.set_property("public", False)
                goep.set_property("gridftp", "node:%s-gridftp" % domain_name)
                
                if self.get((domain_name,"go-auth")) == "myproxy":
                    goep.set_property("myproxy", "node:%s-myproxy" % domain_name)
                else:
                    goep.set_property("myproxy", "myproxy.globusonline.org")

                goep.set_property("globus_connect_cert", self.get((domain_name,"go-gc")))
                    
                domain.add_to_array("go_endpoints", goep)
                
        return topology





# -------------------------------------------------------------------------- #
# Copyright 2010-2011, University of Chicago                                 #
#                                                                            #
# Licensed under the Apache License, Version 2.0 (the "License"); you may    #
# not use this file except in compliance with the License. You may obtain    #
# a copy of the License at                                                   #
#                                                                            #
# http://www.apache.org/licenses/LICENSE-2.0                                 #
#                                                                            #
# Unless required by applicable law or agreed to in writing, software        #
# distributed under the License is distributed on an "AS IS" BASIS,          #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.   #
# See the License for the specific language governing permissions and        #
# limitations under the License.                                             #
# -------------------------------------------------------------------------- #

from distribute_setup import use_setuptools
use_setuptools(version="0.6.15")
from setuptools import setup, find_packages
import sys

sys.path.insert(0, './src')
from globus.provision import RELEASE

cmds = {"globus.provision.cli.api":
        ["gp-instance-create", "gp-instance-describe", "gp-instance-start", "gp-instance-update", 
         "gp-instance-stop", "gp-instance-terminate", "gp-instance-list", "gp-instance-add-user", "gp-instance-add-host",
         "gp-instance-remove-users", "gp-instance-remove-hosts"],
         
        "globus.provision.cli.ec2":
        ["gp-ec2-create-ami", "gp-ec2-update-ami"],
        
        "globus.provision.cli.globusonline":
        ["gp-go-register-endpoints"]
        }

eps = []
for mod in cmds:
    for name in cmds[mod]:
        clsname = name.replace("-","_")
        eps.append("%s = %s:%s_func" % (name, mod, clsname))


setup(name='globus-provision',
      version=RELEASE,
      description='A tool for deploying fully-configured Globus systems on Amazon EC2',
      author='University of Chicago',
      author_email='borja@cs.uchicago.edu',
      url='http://globus.org/provision',
      package_dir = {'': 'src'},      
      packages=find_packages("src", exclude=["dg_paraproxy"]),
      
      install_requires = ['boto>=2.0', 'paramiko>=1.7.7.1', 'colorama>=0.2.4', 'pyOpenSSL>=0.10',
                          'globusonline-transfer-api-client>=0.10.7'],
      setup_requires = [ "setuptools_git >= 0.4.2", ],
      include_package_data=True,
      
      entry_points = {
        'console_scripts': eps
        },

      zip_safe = False,

      license="Apache Software License",
      classifiers=[
          'Development Status :: 4 - Beta',
          'Environment :: Console',
          'Intended Audience :: Developers',
          'Intended Audience :: System Administrators',
          'Intended Audience :: Education',
          'License :: OSI Approved :: Apache Software License',
          'Operating System :: POSIX',
          'Programming Language :: Python',
          'Topic :: Scientific/Engineering',
          'Topic :: System :: Distributed Computing'
          ]
     )



import textwrap
import re
import glob
import os.path

cookbooks = ["provision", "globus", "condor", "galaxy"]

def print_section(title, marker):
    print title
    print marker * len(title)

roles_files = glob.glob("../src/globus/provision/chef-files/roles/*.rb")
roles_files.sort()

print_section("Roles", "=")
for role_file in roles_files:
    f = open(role_file)
    role = {}
    for line in f:
        if len(line.strip()) > 0:
            k,v = line.split(" ", 1)
            role[k.strip()] = v.strip().replace("\"","")
        
    print "**%s**" % role["name"]
    print "  %s" % role["description"]
    print
    print "  Runs: ``%s``" % role["run_list"]
    print

for cookbook in cookbooks:
    print_section("``%s`` cookbook" % cookbook, "=")
    print
    
    files = glob.glob("../src/globus/provision/chef-files/cookbooks/%s/recipes/*.rb" % cookbook)
    files.sort()

    for file in files:
        recipe_name = os.path.basename(file).replace(".rb","")
        
        f = open(file)
        
        lines = []
        title = None
        for l in f:
            if l.startswith("## "):
                if l.startswith("## RECIPE:"):
                    title = l.replace("## RECIPE:","").strip()
                else:                
                    lines.append(l.replace("## ","").strip())
        
        description = "\n".join(lines).strip()
        
        if title != None:
            print ".. _chef_%s_%s:" % (cookbook, recipe_name)
            print
            print_section("``%s`` (%s)" % (recipe_name, title), "-")
            print description
            print
    
    print


import globus.provision.cli.api as api
import globus.provision.cli.ec2 as ec2
import globus.provision.cli.globusonline as globusonline
from globus.provision.cli import Command
from globus.provision.common.utils import rest_table

from docutils.core import publish_string
import re
import textwrap
import operator

OPTION_LEN=50
DESCRIPTION_LEN=100

commands = [api.gp_instance_create, api.gp_instance_start, api.gp_instance_describe,
            api.gp_instance_update, api.gp_instance_stop, api.gp_instance_terminate,
            api.gp_instance_add_host, api.gp_instance_add_user, api.gp_instance_remove_hosts,
            api.gp_instance_remove_users, api.gp_instance_list,
            
            ec2.gp_ec2_create_ami, ec2.gp_ec2_update_ami,
            
            globusonline.gp_go_register_endpoints
            ]

def print_section(title, marker):
    print title
    print marker * len(title)

commands.sort(key=operator.attrgetter("name"))
commands.insert(0, Command)
common_options = Command([]).optparser.option_list
common_options = [str(opt) for opt in common_options]

for command in commands:
    c = command([])
    print 
    if command == Command:
        print ".. _cli_common:"
        print
        print_section("Common options", "=")
        print
        print "These are the options common to all Globus Provision commands."
        print
    else:
        print ".. _cli_%s:" % command.name
        print
        print_section("``%s``" % command.name, "=")
        print
        doc = command.__doc__
        if doc != None:
            doc = textwrap.dedent(doc).strip()
        else:
            doc = "TODO"
        print doc
        print

    opts = c.optparser.option_list
    if command != Command:
        opts = [opt for opt in opts if str(opt) not in common_options]
    c.optparser.formatter.store_option_strings(c.optparser)
    
    if len(opts) > 0:
        col_names = ["Option", "Description"]
        rows = []
        
        for opt in opts:
            if opt.action != "help":
                opt_string = "``%s``" % c.optparser.formatter.option_strings[opt]            
                opt_help = textwrap.dedent(opt.help).strip()                
                
                rows.append([opt_string, opt_help])
                
        print rest_table(col_names, rows)
    
    else:
        print "This command has no options (besides the :ref:`common options <cli_common>`)"    



    print
        


# Copyright (C) 2009-2011  Rene Koecher <shirk@bitspin.org>
#
# This file is part of paraproxy.
#
# Paraproxy is free software; you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 2.1 of the License, or (at your option)
# any later version.
#
# Paraproxy is distrubuted in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with paraproxy; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.

import socket

class WrapSock(object):
	"""
	Wrap a unix domain socket into a file-like object.
	"""
	def __init__(self, sock):
		self._sock = sock
		
	def __getattribute__(self, attr):
#		print 'getattr %s' % attr
		try:
			return super(WrapSock, self).__getattribute__(attr)
		except AttributeError:
			return super(WrapSock, self).__getattribute__('_sock').__getattribute__(attr)
	
	def dup(self):
		return WrapSock(self._sock.dup())
	
	def read(self, *args):
		return self._sock.recv(*args)
	
	def write(self, *args):
		return self._sock.send(*args)




# -------------------------------------------------------------------------- #
# Copyright 2010-2011, University of Chicago                                 #
#                                                                            #
# Licensed under the Apache License, Version 2.0 (the "License"); you may    #
# not use this file except in compliance with the License. You may obtain    #
# a copy of the License at                                                   #
#                                                                            #
# http://www.apache.org/licenses/LICENSE-2.0                                 #
#                                                                            #
# Unless required by applicable law or agreed to in writing, software        #
# distributed under the License is distributed on an "AS IS" BASIS,          #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.   #
# See the License for the specific language governing permissions and        #
# limitations under the License.                                             #
# -------------------------------------------------------------------------- #

"""
Thread management

This module provides an abstraction over the threading package, including
the ability to manage the execution of a tree of threads.
"""

import os
import sys
import signal
import threading
import traceback
from globus.provision.common import log

class ThreadAbortException(Exception):
    pass
        
class GPThread (threading.Thread):
    def __init__ (self, multi, name, depends = None):
        threading.Thread.__init__(self)
        self.multi = multi
        self.name = name
        self.exception = None
        self.stack_trace = None
        self.status = -1
        self.depends = depends
        
    def check_continue(self):
        if self.multi.abort.is_set():
            raise ThreadAbortException()
        
    def run2(self):
        pass
        
    def run(self):
        try:
            self.run2()
            self.status = 0
        except Exception:
            exc_type, exc_value, exc_traceback = sys.exc_info()
            self.exception = exc_value
            self.stack_trace = traceback.format_exception(exc_type, exc_value, exc_traceback)
            self.status = 1
            self.multi.thread_failure(self)
            
        if self.status == 0:
            self.multi.thread_success(self)

class MultiThread(object):
    def __init__(self):
        self.num_threads = 0
        self.done_threads = 0
        self.threads = {}
        self.lock = threading.Lock()
        self.all_done = threading.Event()
        self.abort = threading.Event()

    def add_thread(self, thread):
        self.threads[thread.name] = thread     
        self.num_threads += 1

    def run(self):
        self.done_threads = 0
        for t in [th for th in self.threads.values() if th.depends == None]:
            t.start()
        self.all_done.wait()
        
    def thread_success(self, thread):
        with self.lock:
            self.done_threads += 1
            log.debug("%s thread has finished successfully." % thread.name)
            log.debug("%i threads are done. Remaining: %s" % (self.done_threads, ",".join([t.name for t in self.threads.values() if t.status == -1])))
            for t in [th for th in self.threads.values() if th.depends == thread]:
                t.start()            
            if self.done_threads == self.num_threads:
                self.all_done.set()            

    def thread_failure(self, thread):
        with self.lock:
            if not isinstance(thread.exception, ThreadAbortException):
                log.debug("%s thread has failed: %s" % (thread.name, thread.exception))
                self.abort.set()
            else:
                log.debug("%s thread is being aborted." % thread.name)
                thread.status = 2
            self.done_threads += 1
            self.abort_dependents(thread)
            log.debug("%i threads are done. Remaining: %s" % (self.done_threads, ",".join([t.name for t in self.threads.values() if t.status == -1])))
            if self.done_threads == self.num_threads:
                self.all_done.set()           
                
    def abort_dependents(self, thread):
        dep = [th for th in self.threads.values() if th.depends == thread]
        for th in dep:
            log.debug("%s thread is being aborted because it depends on failed %s thread." % (th.name, thread.name))
            th.status = 3
            self.done_threads += 1
            self.abort_dependents(th)
        
    def all_success(self):
        return all([t.status == 0 for t in self.threads.values()])
        
    def get_exceptions(self):
        return dict([(t.name, (t.exception, t.stack_trace)) for t in self.threads.values() if t.status == 1]) 

# From http://code.activestate.com/recipes/496735-workaround-for-missed-sigint-in-multithreaded-prog/
# Modified so it will run a cleanup function
class SIGINTWatcher(object):
    """this class solves two problems with multithreaded
    programs in Python, (1) a signal might be delivered
    to any thread (which is just a malfeature) and (2) if
    the thread that gets the signal is waiting, the signal
    is ignored (which is a bug).

    The watcher is a concurrent process (not thread) that
    waits for a signal and the process that contains the
    threads.  See Appendix A of The Little Book of Semaphores.
    http://greenteapress.com/semaphores/

    I have only tested this on Linux.  I would expect it to
    work on the Macintosh and not work on Windows.
    """
    
    def __init__(self, cleanup_func):
        """ Creates a child thread, which returns.  The parent
            thread waits for a KeyboardInterrupt and then kills
            the child thread.
        """
        self.cleanup_func = cleanup_func
        self.child = os.fork()
        if self.child == 0:
            return
        else:
            self.watch()

    def watch(self):
        try:
            os.wait()
        except KeyboardInterrupt:
            self.cleanup_func()
            self.kill()
        sys.exit()

    def kill(self):
        try:
            os.kill(self.child, signal.SIGKILL)
        except OSError: pass



from globus.provision.core.config import SimpleTopologyConfig
import tempfile
import shutil

def load_config_file(f, dummy):
    configf = open(f)
    config_txt = configf.read()
    if dummy:
        config_txt = config_txt.replace("deploy: ec2", "deploy: dummy")
    configf.close()

    topology_file = f
    conf = SimpleTopologyConfig(topology_file)
    topology = conf.to_topology()
    topology_json = topology.to_json_string()   
    
    return config_txt, topology_json

def create_temp_config_file(f, tempf, dummy):
    configf = open(f)
    config_txt = configf.read()
    if dummy:
        config_txt = config_txt.replace("deploy: ec2", "deploy: dummy")
    configf.close()
    
    configf = open(tempf, "w")
    configf.write(config_txt)
    configf.close()

def create_instances_dir():
    return tempfile.mkdtemp(prefix="gptesttmp")

def remove_instances_dir(d):
    shutil.rmtree(d)


from globus.provision.common.utils import DemoGridThread, MultiThread, SIGINTWatcher, SSH
import sys

class SSHTestThread(DemoGridThread):
    def __init__(self, multi, name, fail = 0, depends = None):
        DemoGridThread.__init__(self, multi, name, depends)
        
    def run2(self):
        ssh_out = sys.stdout
        ssh_err = sys.stderr
        raise Exception("FAIL")
        print "%s - Establishing SSH connection" % self.name
        ssh = SSH("borja", "polaris.cs.uchicago.edu", "/home/borja/.ssh/id_rsa", ssh_out, ssh_err)
        ssh.open()
        print "%s - SSH connection established" % self.name
        
        ssh.run("date; /bin/hostname")
        
def killed():
    print "Killed"
    
SIGINTWatcher(killed) 

mt = MultiThread()

t1 = SSHTestThread(mt, "T1", fail = 0)
t2 = SSHTestThread(mt, "T2", fail = 0)
t3 = SSHTestThread(mt, "T3", fail = 0, depends = t1)

print "foo"
mt.add_thread(t1)
mt.add_thread(t2)
mt.add_thread(t3)
print "bar"        
mt.run()
print mt.all_success()

for th in mt.threads.values():
    print th.name, th.status

